import polars as pl
import pytest
from polars.testing import assert_frame_equal
from pyarrow import Table

from matchbox.client.results import Results


def test_clusters_and_root_leaf():
    """From a results object, we can derive clusters at various levels."""
    # Prepare dummy data and model
    left_root_leaf = pl.DataFrame(
        [
            # Two keys per root
            {"id": 10, "leaf_id": 1},
            {"id": 10, "leaf_id": 1},
            # Two leaves per root (same representation)
            {"id": 20, "leaf_id": 2},
            {"id": 20, "leaf_id": 3},
            # Singleton cluster with two keys
            {"id": 4, "leaf_id": 4},
            {"id": 4, "leaf_id": 4},
        ]
    )

    right_root_leaf = pl.DataFrame(
        # For simplicity, all these are singleton clusters
        [
            {"id": 5, "leaf_id": 5},
            {"id": 6, "leaf_id": 6},
            {"id": 7, "leaf_id": 7},
            {"id": 8, "leaf_id": 8},
        ]
    )

    probabilities = Table.from_pylist(
        [
            # simple left-right merge
            {"left_id": 4, "right_id": 5, "probability": 100},
            # dedupe through linking
            {"left_id": 10, "right_id": 6, "probability": 100},
            {"left_id": 10, "right_id": 7, "probability": 100},
        ]
    )

    results = Results(
        probabilities=probabilities,
        left_root_leaf=left_root_leaf.to_arrow(),
        right_root_leaf=right_root_leaf.to_arrow(),
    )

    # Check two ways of representing clusters
    clusters = pl.from_arrow(results.clusters)
    grouped_children = {
        tuple(sorted(group))
        for group in clusters.group_by("parent").agg("child")["child"]
    }
    # Only the input IDs referenced in probabilities are present, in the right groups
    assert grouped_children == {(4, 5), (6, 7, 10)}

    root_leaf = results.root_leaf()
    grouped_leaves = {
        tuple(sorted(group))
        for group in root_leaf.group_by("root_id").agg("leaf_id")["leaf_id"]
    }
    # Only single-digits are present, and all of them, in the right groups
    assert grouped_leaves == {(1, 6, 7), (2, 3), (4, 5), (8,)}

    # To check edge cases, look at no probabilities returned
    empty_results = Results(
        probabilities=Table.from_pydict(
            {"left_id": [], "right_id": [], "probability": []}
        ),
        left_root_leaf=left_root_leaf.to_arrow(),
        right_root_leaf=right_root_leaf.to_arrow(),
    )

    assert len(empty_results.clusters) == 0
    expected_empty_root_leaf = pl.concat(
        [
            left_root_leaf.rename({"id": "root_id"}),
            right_root_leaf.rename({"id": "root_id"}),
        ]
    ).unique()
    assert_frame_equal(
        empty_results.root_leaf(),
        expected_empty_root_leaf,
        check_column_order=False,
        check_row_order=False,
    )

    # The above was only possible because leaf IDs were present in the inputs
    only_prob_results = Results(probabilities=probabilities)
    with pytest.raises(RuntimeError, match="instantiated for validation"):
        only_prob_results.root_leaf()



"""Resolution API routes for the Matchbox server."""

from typing import Annotated

from fastapi import APIRouter, Body, Depends, HTTPException, Query, status

from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import (
    BackendResourceType,
    CRUDOperation,
    NotFoundError,
    Resolution,
    ResolutionOperationStatus,
    UploadStatus,
)
from matchbox.common.exceptions import (
    MatchboxDeletionNotConfirmed,
    MatchboxResolutionNotFoundError,
)
from matchbox.common.graph import ModelResolutionName, ResolutionName, ResolutionType
from matchbox.server.api.dependencies import (
    BackendDependency,
    ParquetResponse,
    UploadTrackerDependency,
    authorisation_dependencies,
)

router = APIRouter(prefix="/resolutions", tags=["resolution"])


@router.post(
    "",
    responses={
        500: {
            "model": ResolutionOperationStatus,
            **ResolutionOperationStatus.status_500_examples(),
        },
    },
    status_code=status.HTTP_201_CREATED,
    dependencies=[Depends(authorisation_dependencies)],
)
def create_resolution(
    backend: BackendDependency,
    resolution: Resolution,
) -> ResolutionOperationStatus | UploadStatus:
    """Create a resolution (model or source)."""
    try:
        backend.insert_resolution(resolution=resolution)
        return ResolutionOperationStatus(
            success=True,
            name=resolution.name,
            operation=CRUDOperation.CREATE,
        )
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResolutionOperationStatus(
                success=False,
                name=resolution.name,
                operation=CRUDOperation.CREATE,
                details=str(e),
            ).model_dump(),
        ) from e


@router.get(
    "/{name}",
    responses={404: {"model": NotFoundError}},
)
def get_resolution(
    backend: BackendDependency,
    name: ResolutionName,
    validate_type: ResolutionType | None = None,
) -> Resolution:
    """Get a resolution (model or source) from the backend."""
    try:
        return backend.get_resolution(name=name, validate=validate_type)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e


@router.get(
    "/{name}/sources",
    responses={404: {"model": NotFoundError}},
)
def get_leaf_source_resolutions(
    backend: BackendDependency,
    name: ResolutionName,
) -> list[Resolution]:
    """Get all sources in scope for a resolution."""
    try:
        return backend.get_leaf_source_resolutions(name=name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e


@router.delete(
    "/{name}",
    responses={
        404: {"model": NotFoundError},
        409: {
            "model": ResolutionOperationStatus,
            **ResolutionOperationStatus.status_409_examples(),
        },
    },
    dependencies=[Depends(authorisation_dependencies)],
)
def delete_resolution(
    backend: BackendDependency,
    name: ResolutionName,
    certain: Annotated[
        bool, Query(description="Confirm deletion of the resolution")
    ] = False,
) -> ResolutionOperationStatus:
    """Delete a resolution from the backend."""
    try:
        backend.delete_resolution(name=name, certain=certain)
        return ResolutionOperationStatus(
            success=True,
            name=name,
            operation=CRUDOperation.DELETE,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e
    except MatchboxDeletionNotConfirmed as e:
        raise HTTPException(
            status_code=409,
            detail=ResolutionOperationStatus(
                success=False,
                name=name,
                operation=CRUDOperation.DELETE,
                details=str(e),
            ).model_dump(),
        ) from e


@router.post(
    "/{name}/data",
    responses={404: {"model": NotFoundError}},
    status_code=status.HTTP_202_ACCEPTED,
    dependencies=[Depends(authorisation_dependencies)],
)
def set_data(
    backend: BackendDependency,
    upload_tracker: UploadTrackerDependency,
    name: ResolutionName,
    validate_type: ResolutionType | None = None,
) -> UploadStatus:
    """Create an upload task for source hashes or model results."""
    try:
        resolution = backend.get_resolution(name=name, validate=validate_type)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e

    if resolution.resolution_type == ResolutionType.SOURCE:
        upload_id = upload_tracker.add_source(metadata=resolution)
        return upload_tracker.get(upload_id=upload_id).status

    upload_id = upload_tracker.add_model(metadata=resolution)
    return upload_tracker.get(upload_id=upload_id).status


@router.get(
    "/{name}/data",
    responses={404: {"model": NotFoundError}},
)
def get_results(
    backend: BackendDependency, name: ModelResolutionName
) -> ParquetResponse:
    """Download results for a model as a parquet file."""
    try:
        res = backend.get_model_data(name=name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e

    buffer = table_to_buffer(res)
    return ParquetResponse(buffer.getvalue())


@router.patch(
    "/{name}/truth",
    responses={
        404: {"model": NotFoundError},
        500: {
            "model": ResolutionOperationStatus,
            **ResolutionOperationStatus.status_500_examples(),
        },
    },
    dependencies=[Depends(authorisation_dependencies)],
)
def set_truth(
    backend: BackendDependency,
    name: ModelResolutionName,
    truth: Annotated[int, Body(ge=0, le=100)],
) -> ResolutionOperationStatus:
    """Set truth data for a resolution."""
    try:
        # This will fail for a source, which is what we want for now
        backend.set_model_truth(name=name, truth=truth)
        return ResolutionOperationStatus(
            success=True,
            name=name,
            operation=CRUDOperation.UPDATE,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e
    except Exception as e:
        raise HTTPException(
            status_code=500,
            detail=ResolutionOperationStatus(
                success=False,
                name=name,
                operation=CRUDOperation.UPDATE,
                details=str(e),
            ).model_dump(),
        ) from e


@router.get(
    "/{name}/truth",
    responses={404: {"model": NotFoundError}},
)
def get_truth(backend: BackendDependency, name: ModelResolutionName) -> float:
    """Get truth data for a resolution."""
    try:
        return backend.get_model_truth(name=name)
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e



"""Common database utilities for Matchbox."""

from collections.abc import Callable, Iterator
from enum import StrEnum
from typing import (
    TYPE_CHECKING,
    Any,
    Literal,
    TypeAlias,
    TypeVar,
    overload,
)

import polars as pl
import sqlglot
from pandas import DataFrame as PandasDataFrame
from polars import DataFrame as PolarsDataFrame
from pyarrow import Table as ArrowTable
from sqlalchemy.engine import Engine

from matchbox.common.exceptions import MatchboxSourceExtractTransformError

if TYPE_CHECKING:
    from adbc_driver_postgresql.dbapi import Connection as ADBCConnection
else:
    ADBCConnection = Any


class QueryReturnType(StrEnum):
    """Enumeration of dataframe types to return from query."""

    PANDAS = "pandas"
    POLARS = "polars"
    ARROW = "arrow"


QueryReturnClass: TypeAlias = ArrowTable | PandasDataFrame | PolarsDataFrame

T = TypeVar("T")


@overload
def sql_to_df(
    stmt: str,
    connection: Engine | ADBCConnection,
    return_type: QueryReturnType,
    *,
    return_batches: Literal[False] = False,
    batch_size: int | None = None,
    rename: dict[str, str] | Callable | None = None,
    schema_overrides: dict[str, pl.DataType] | None = None,
    execute_options: dict[str, Any] | None = None,
) -> QueryReturnClass: ...


@overload
def sql_to_df(
    stmt: str,
    connection: Engine | ADBCConnection,
    return_type: QueryReturnType,
    *,
    return_batches: Literal[True],
    batch_size: int | None = None,
    rename: dict[str, str] | Callable | None = None,
    schema_overrides: dict[str, pl.DataType] | None = None,
    execute_options: dict[str, Any] | None = None,
) -> Iterator[QueryReturnClass]: ...


def sql_to_df(
    stmt: str,
    connection: Engine | ADBCConnection,
    return_type: QueryReturnType = QueryReturnType.PANDAS,
    *,
    return_batches: bool = False,
    batch_size: int | None = None,
    rename: dict[str, str] | Callable | None = None,
    schema_overrides: dict[str, pl.DataType] | None = None,
    execute_options: dict[str, Any] | None = None,
) -> QueryReturnClass | Iterator[QueryReturnClass]:
    """Executes the given SQLAlchemy statement or SQL string using Polars.

    Args:
        stmt: A SQL string to be executed.
        connection: A SQLAlchemy Engine object or
            ADBC connection.
        return_type: The type of the return value. One of "arrow", "pandas",
            or "polars".
        return_batches: If True, return an iterator that yields each batch
            separately. If False, return a single DataFrame with all results.
            Default is False.
        batch_size: Indicate the size of each batch when processing
            data in batches. Default is None.
        rename: A dictionary mapping old column
            names to new column names, or a callable that takes a DataFrame and
            returns a DataFrame with renamed columns. Default is None.
        schema_overrides: A dictionary mapping column names
            to dtypes. Default is None.
        execute_options: These options will be passed through
            into the underlying query execution method as kwargs. Default is None.

    Returns:
        If return_batches is False: A dataframe of the query results in the specified
            format.
        If return_batches is True: An iterator of dataframes in the specified format.

    Raises:
        ValueError:

            * If the connection is not properly configured or if an unsupported
                return type is specified.
            * If batch_size and return_batches are either both set or both unset.

    """
    if not batch_size and return_batches:
        raise ValueError("A batch size must be specified if return_batches is True")

    if batch_size and not return_batches:
        raise ValueError("Cannot set a batch size if return_batches if False")

    def _to_format(results: PolarsDataFrame) -> QueryReturnClass:
        """Convert the results to the specified format."""
        if rename:
            results = results.rename(rename)

        match return_type:
            case QueryReturnType.POLARS:
                return results
            case QueryReturnType.PANDAS:
                return results.to_pandas()
            case QueryReturnType.ARROW:
                return results.to_arrow()
            case _:
                raise ValueError("Unknown return type specified.")

    res = pl.read_database(
        query=stmt,
        connection=connection,
        iter_batches=bool(batch_size),
        batch_size=batch_size,
        schema_overrides=schema_overrides,
        execute_options=execute_options,
    )

    if return_batches:
        return (_to_format(batch) for batch in res)

    return _to_format(res)


def validate_sql_for_data_extraction(sql: str) -> bool:
    """Validates that the SQL statement only contains a single data-extracting command.

    Args:
        sql: The SQL statement to validate

    Returns:
        bool: True if the SQL statement is valid

    Raises:
        ParseError: If the SQL statement cannot be parsed
        MatchboxSourceExtractTransformError: If validation requirements are not met
    """
    if not sql.strip():
        raise MatchboxSourceExtractTransformError(
            "SQL statement is empty or only contains whitespace."
        )

    expressions = sqlglot.parse(sql)

    if len(expressions) > 1:
        raise MatchboxSourceExtractTransformError(
            "SQL statement contains multiple commands."
        )

    if not expressions:
        raise MatchboxSourceExtractTransformError(
            "SQL statement does not contain any valid expressions."
        )

    expr = expressions[0]

    if not isinstance(expr, sqlglot.expressions.Select):
        raise MatchboxSourceExtractTransformError(
            "SQL statement must start with a SELECT or WITH command."
        )

    forbidden = (
        sqlglot.expressions.DDL,
        sqlglot.expressions.DML,
        sqlglot.expressions.Into,
    )

    if len(list(expr.find_all(forbidden))) > 0:
        raise MatchboxSourceExtractTransformError(
            "SQL statement must not contain DDL or DML commands."
        )

    return True



import pytest
from rustworkx import PyDiGraph

from matchbox.common.graph import (
    ResolutionEdge,
    ResolutionGraph,
    ResolutionNode,
)
from matchbox.common.graph import (
    ResolutionType as ResType,
)


@pytest.fixture
def resolution_graph() -> ResolutionGraph:
    res_graph = ResolutionGraph(
        nodes={
            ResolutionNode(id=1, name="1", type=ResType.SOURCE),
            ResolutionNode(id=2, name="2", type=ResType.SOURCE),
            ResolutionNode(id=3, name="3", type=ResType.MODEL),
            ResolutionNode(id=4, name="4", type=ResType.MODEL),
            ResolutionNode(id=5, name="5", type=ResType.MODEL),
        },
        edges={
            ResolutionEdge(parent=1, child=3),
            ResolutionEdge(parent=2, child=5),
            ResolutionEdge(parent=3, child=4),
            ResolutionEdge(parent=4, child=5),
        },
    )

    return res_graph


@pytest.fixture
def pydigraph() -> PyDiGraph:
    G = PyDiGraph()
    n1 = G.add_node({"id": 1, "name": "1", "type": str(ResType.SOURCE)})
    n2 = G.add_node({"id": 2, "name": "2", "type": str(ResType.SOURCE)})
    n3 = G.add_node({"id": 3, "name": "3", "type": str(ResType.MODEL)})
    n4 = G.add_node({"id": 4, "name": "4", "type": str(ResType.MODEL)})
    n5 = G.add_node({"id": 5, "name": "5", "type": str(ResType.MODEL)})
    G.add_edge(n1, n3, {})
    G.add_edge(n2, n5, {})
    G.add_edge(n3, n4, {})
    G.add_edge(n4, n5, {})
    return G



"""Objects to define a DAG which indexes, deduplicates and links data."""

import datetime
from collections import defaultdict
from typing import Self

from pyarrow import Table as ArrowTable

from matchbox.client import _handler
from matchbox.client.models import Model
from matchbox.client.queries import Query
from matchbox.client.sources import Source
from matchbox.common.exceptions import MatchboxResolutionNotFoundError
from matchbox.common.logging import logger


class DAG:
    """Self-sufficient pipeline of indexing, deduping and linking steps."""

    def __init__(self, name: str, new: bool = False):
        """Initialises empty DAG."""
        self.name = name
        self.nodes: dict[str, Source | Model] = {}
        self.graph: dict[str, list[str]] = {}

    def _check_dag(self, dag: Self):
        if self != dag:
            raise ValueError("Cannot mix DAGs")

    def _add_step(self, step: Source | Model) -> None:
        """Validate and add sources and models to DAG."""
        self._check_dag(step.dag)

        if step.name in self.nodes:
            raise ValueError(f"Name '{step.name}' is already taken in the DAG")
        if isinstance(step, Model):
            self._check_dag(step.left_query.dag)
            if step.right_query:
                self._check_dag(step.right_query.dag)

            for resolution_name in step.dependencies:
                if resolution_name not in self.nodes:
                    raise ValueError(f"Step {resolution_name} not added to DAG")
            self.graph[step.name] = step.parents
        else:
            self.graph[step.name] = []

        self.nodes[step.name] = step

    @property
    def final_step(self) -> str:
        """Returns name of the root node in the DAG.

        Returns:
            The name of the root node in the DAG

        Raises:
            ValueError: If the DAG does not have a final step
        """
        if not self.nodes:
            raise ValueError("The DAG is empty.")

        inverse_graph = defaultdict(list)
        for node in self.graph:
            for neighbor in self.graph[node]:
                inverse_graph[neighbor].append(node)

        apex_nodes = {node for node in self.graph if node not in inverse_graph}
        if len(apex_nodes) > 1:
            raise ValueError("Some models or sources are disconnected")
        elif not apex_nodes:
            raise ValueError("No root node found, DAG might contain cycles")
        else:
            return apex_nodes.pop()

    def source(self, *args, **kwargs) -> Source:
        """Create Source and add it to the DAG."""
        source = Source(*args, **kwargs, dag=self)
        self._add_step(source)

        return source

    def model(self, *args, **kwargs) -> Model:
        """Create Model and add it to the DAG."""
        model = Model(*args, **kwargs, dag=self)
        self._add_step(model)

        return model

    def query(self, *args, **kwargs) -> Query:
        """Create Query object."""
        return Query(*args, **kwargs, dag=self)

    def draw(
        self,
        start_time: datetime.datetime | None = None,
        doing: str | None = None,
        skipped: list[str] | None = None,
    ) -> str:
        """Create a string representation of the DAG as a tree structure.

        If `start_time` is provided, it will show the status of each node
        based on the last run time. The status indicators are:

        * ✅ Done
        * 🔄 Working
        * ⏸️ Awaiting
        * ⏭️ Skipped

        Args:
            start_time: Start time of the DAG run. Used to calculate node status.
            doing: Name of the node currently being processed (if any).
            skipped: List of node names that were skipped.

        Returns:
            String representation of the DAG with status indicators.
        """
        root_name = self.final_step
        skipped = skipped or []

        def _get_node_status(name: str) -> str:
            """Determine the status indicator for a node."""
            if name in skipped:
                return "⏭️"
            elif doing and name == doing:
                return "🔄"
            elif (
                (node := self.nodes.get(name))
                and node.last_run
                and node.last_run > start_time
            ):
                return "✅"
            else:
                return "⏸️"

        # Add status indicator if start_time is provided
        if start_time is not None:
            status = _get_node_status(root_name)
            result = [f"{status} {root_name}"]
        else:
            result = [root_name]

        visited = set([root_name])

        def format_children(node: str, prefix=""):
            """Recursively format the children of a node."""
            children = []
            # Get all outgoing edges from this node
            for target in self.graph.get(node, []):
                if target not in visited:
                    children.append(target)
                    visited.add(target)

            # Format each child
            for i, child in enumerate(children):
                is_last = i == len(children) - 1

                # Add status indicator if start_time is provided
                if start_time is not None:
                    status = _get_node_status(child)
                    child_display = f"{status} {child}"
                else:
                    child_display = child

                if is_last:
                    result.append(f"{prefix}└── {child_display}")
                    format_children(child, prefix + "    ")
                else:
                    result.append(f"{prefix}├── {child_display}")
                    format_children(child, prefix + "│   ")

        format_children(root_name)

        return "\n".join(result)

    def run_and_sync(
        self,
        full_rerun: bool = False,
        start: str | None = None,
        finish: str | None = None,
    ):
        """Run entire DAG and send results to server."""
        start_time = datetime.datetime.now()

        # Determine order of execution steps
        root_node = self.final_step

        def depth_first(node: str, sequence: list):
            sequence.append(node)
            for neighbour in self.graph[node]:
                if neighbour not in sequence:
                    depth_first(neighbour, sequence)

        inverse_sequence = []
        depth_first(root_node, inverse_sequence)
        sequence = list(reversed(inverse_sequence))

        # Identify skipped nodes
        skipped_nodes = []
        if start:
            try:
                start_index = sequence.index(start)
                skipped_nodes = sequence[:start_index]
            except ValueError as e:
                raise ValueError(f"Step {start} not in DAG") from e
        else:
            start_index = 0

        # Determine end index
        if finish:
            try:
                end_index = sequence.index(finish) + 1
                skipped_nodes.extend(sequence[end_index:])
            except ValueError as e:
                raise ValueError(f"Step {finish} not in DAG") from e
        else:
            end_index = len(sequence)

        sequence = sequence[start_index:end_index]
        if not full_rerun:
            # Exclude nodes that already run, unless a full re-run is forced
            already_run = [node for node in sequence if self.nodes[node].last_run]
            skipped_nodes.extend(already_run)
            sequence = [node for node in sequence if node not in already_run]

        for step_name in sequence:
            node = self.nodes[step_name]

            logger.info(
                "\n"
                + self.draw(
                    start_time=start_time, doing=node.name, skipped=skipped_nodes
                )
            )
            try:
                node.run(full_rerun=full_rerun)
                node.sync()
            except Exception as e:
                logger.error(f"❌ {node.name} failed: {e}")
                raise e

        logger.info("\n" + self.draw(start_time=start_time, skipped=skipped_nodes))

    def lookup_key(
        self,
        from_source: str,
        to_sources: list[str],
        key: str,
        threshold: int | None = None,
    ) -> dict[str, list[str]]:
        """Matches IDs against the selected backend.

        Args:
            from_source: Name of source the provided key belongs to
            to_sources: Names of sources to find keys in
            key: The value to match from the source. Usually a primary key
            threshold (optional): The threshold to use for creating clusters.
                If None, uses the resolutions' default threshold
                If an integer, uses that threshold for the specified resolution, and the
                resolution's cached thresholds for its ancestors

        Examples:
            ```python
            dag.lookup_key(
                from_source="companies_house",
                to_sources=[
                    "datahub_companies",
                    "hmrc_exporters",
                ]
                key="8534735",
            )
            ```
        """
        matches = _handler.match(
            targets=to_sources,
            source=from_source,
            key=key,
            resolution=self.final_step,
            threshold=threshold,
        )

        to_sources_results = {m.target: list(m.target_id) for m in matches}
        # If no matches, _handler will raise
        return {from_source: list(matches[0].source_id), **to_sources_results}

    def extract_lookup(
        self,
        source_filter: list[str] | None = None,
        location_names: list[str] | None = None,
    ) -> ArrowTable:
        """Return matchbox IDs to source key mapping, optionally filtering.

        Args:
            source_filter: An optional list of source resolution names to filter by.
            location_names: An optional list of location names to filter by.
        """
        # Get all sources in scope of the resolution
        point_of_truth = self.final_step
        source_resolutions = _handler.get_leaf_source_resolutions(name=point_of_truth)

        if source_filter:
            source_resolutions = [
                s for s in source_resolutions if s.name in source_filter
            ]

        if location_names:
            source_resolutions = [
                s
                for s in source_resolutions
                if s.config.location_config.name in location_names
            ]

        if not source_resolutions:
            raise MatchboxResolutionNotFoundError("No compatible source was found")

        source_mb_ids: list[ArrowTable] = []
        source_to_key_field: dict[str, str] = {}

        for s in source_resolutions:
            # Get Matchbox IDs from backend
            source_mb_ids.append(
                _handler.query(
                    source=s.name, resolution=point_of_truth, return_leaf_id=False
                )
            )

            source_to_key_field[s.name] = s.config.key_field.name

        # Join Matchbox IDs to form mapping table
        mapping = source_mb_ids[0]
        mapping = mapping.rename_columns(
            {
                "key": source_resolutions[0].config.qualified_key(
                    source_resolutions[0].name
                )
            }
        )
        if len(source_resolutions) > 1:
            for s, mb_ids in zip(
                source_resolutions[1:], source_mb_ids[1:], strict=True
            ):
                mapping = mapping.join(
                    right_table=mb_ids, keys="id", join_type="full outer"
                )
                mapping = mapping.rename_columns(
                    {"key": s.config.qualified_key(s.name)}
                )

        return mapping



"""Objects representing the results of running a model client-side."""

from collections.abc import Hashable
from typing import ParamSpec, TypeVar

import polars as pl
import pyarrow as pa
import pyarrow.compute as pc
from pydantic import ConfigDict

from matchbox.common.arrow import SCHEMA_RESULTS
from matchbox.common.hash import IntMap
from matchbox.common.transform import to_clusters

T = TypeVar("T", bound=Hashable)
P = ParamSpec("P")
R = TypeVar("R")


class Results:
    """Results of a model run.

    Contains:

    * The probabilities of each pair being a match
    * (Optional) The clusters of connected components at each threshold
    * (Optional) The input data to the model that generated the probabilities


    Allows users to easily interrogate the outputs of models, explore decisions on
    choosing thresholds for clustering, and upload the results to Matchbox.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)

    probabilities: pa.Table
    _clusters: pa.Table | None = None

    def __init__(
        self,
        probabilities: pa.Table | pl.DataFrame,
        left_root_leaf: pa.Table | None = None,
        right_root_leaf: pa.Table | None = None,
    ) -> None:
        """Initialises and validates results."""
        self.left_root_leaf = None
        self.right_root_leaf = None

        if left_root_leaf is not None:
            self.left_root_leaf = left_root_leaf
        if right_root_leaf is not None:
            self.right_root_leaf = right_root_leaf

        if isinstance(probabilities, pl.DataFrame):
            probabilities = probabilities.to_arrow()

        if not isinstance(probabilities, pa.Table):
            raise ValueError("Expected a polars DataFrame or pyarrow Table.")

        expected_fields = set(SCHEMA_RESULTS.names)
        if set(probabilities.column_names) != expected_fields:
            raise ValueError(
                f"Expected {expected_fields}.\nFound {set(probabilities.column_names)}."
            )

        # Handle empty tables
        if probabilities.num_rows == 0:
            empty_arrays = [pa.array([], type=field.type) for field in SCHEMA_RESULTS]
            probabilities = pa.Table.from_arrays(
                empty_arrays, names=[field.name for field in SCHEMA_RESULTS]
            )

        # Process probability field if it contains floating-point or decimal values
        probability_type = probabilities["probability"].type
        if pa.types.is_floating(probability_type) or pa.types.is_decimal(
            probability_type
        ):
            probability_uint8 = pc.cast(
                pc.round(pc.multiply(probabilities["probability"], 100)),
                options=pc.CastOptions(
                    target_type=pa.uint8(),
                    allow_float_truncate=True,
                    allow_decimal_truncate=True,
                ),
            )

            # Check max value only if the table is not empty
            max_prob = pc.max(probability_uint8)
            if max_prob is not None and max_prob.as_py() > 100:
                p_max = pc.max(probabilities["probability"]).as_py()
                p_min = pc.min(probabilities["probability"]).as_py()
                raise ValueError(f"Probability range misconfigured: [{p_min}, {p_max}]")

            probabilities = probabilities.set_column(
                i=probabilities.schema.get_field_index("probability"),
                field_="probability",
                column=probability_uint8,
            )

        self.probabilities = probabilities.cast(SCHEMA_RESULTS)

    @property
    def clusters(self):
        """Retrieve new clusters implied by these results."""
        if not self._clusters:
            im = IntMap()
            self._clusters = to_clusters(
                results=self.probabilities, dtype=pa.int64, hash_func=im.index
            )
        return self._clusters

    def _merge_with_source_data(
        self,
        base_df: pl.DataFrame,
        base_df_cols: list[str],
        left_data: pl.DataFrame,
        left_key: str,
        right_data: pl.DataFrame,
        right_key: str,
        left_merge_col: str,
        right_merge_col: str,
    ) -> pl.DataFrame:
        """Helper method to merge results with source data frames."""
        return (
            base_df.select(base_df_cols)
            .join(
                left_data,
                how="left",
                left_on=left_merge_col,
                right_on=left_key,
            )
            .join(
                right_data,
                how="left",
                left_on=right_merge_col,
                right_on=right_key,
            )
        )

    def inspect_probabilities(
        self,
        left_data: pl.DataFrame,
        left_key: str,
        right_data: pl.DataFrame,
        right_key: str,
    ) -> pl.DataFrame:
        """Enriches the probability results with the source data."""
        return self._merge_with_source_data(
            base_df=pl.from_arrow(self.probabilities),
            base_df_cols=["left_id", "right_id", "probability"],
            left_data=left_data,
            left_key=left_key,
            right_data=right_data,
            right_key=right_key,
            left_merge_col="left_id",
            right_merge_col="right_id",
        )

    def inspect_clusters(
        self,
        left_data: pl.DataFrame,
        left_key: str,
        right_data: pl.DataFrame,
        right_key: str,
    ) -> pl.DataFrame:
        """Enriches the cluster results with the source data."""
        return self._merge_with_source_data(
            base_df=pl.from_arrow(self.clusters),
            base_df_cols=["parent", "child", "probability"],
            left_data=left_data,
            left_key=left_key,
            right_data=right_data,
            right_key=right_key,
            left_merge_col="child",
            right_merge_col="child",
        )

    def root_leaf(self):
        """Returns all roots and leaves implied by these results."""
        if self.left_root_leaf is None:
            raise RuntimeError(
                "This Results object wasn't instantiated for validation features."
            )

        parents_root_leaf = pl.from_arrow(self.left_root_leaf.select(["id", "leaf_id"]))
        if self.right_root_leaf is not None:
            parents_root_leaf = pl.concat(
                [
                    parents_root_leaf,
                    pl.from_arrow(self.right_root_leaf.select(["id", "leaf_id"])),
                ]
            )

        # Go from parent-child (where child could be the root of another model)
        # to root-leaf, where leaf is a source cluster ID
        root_leaf_res = (
            pl.from_arrow(self.clusters)
            .rename({"parent": "root_id"})
            .join(parents_root_leaf, left_on="child", right_on="id")
            .select(["root_id", "leaf_id"])
            .unique()
        )

        # Generate root-leaf for those input rows that weren't merged by this model
        unmerged_ids_rows = (
            parents_root_leaf.select("id", "leaf_id")
            .join(
                pl.from_arrow(self.clusters).select("child"),
                left_on="id",
                right_on="child",
                how="anti",
            )
            .rename({"id": "root_id"})
            .select(["root_id", "leaf_id"])
            .unique()
        )

        return pl.concat([root_leaf_res, unmerged_ids_rows])



"""API routes for the Matchbox server."""

from datetime import datetime
from importlib.metadata import version
from typing import Annotated

from fastapi import (
    BackgroundTasks,
    Depends,
    FastAPI,
    HTTPException,
    Query,
    Request,
    Response,
    UploadFile,
    status,
)
from fastapi.encoders import jsonable_encoder
from fastapi.responses import JSONResponse
from starlette.exceptions import HTTPException as StarletteHTTPException

from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import (
    BackendCountableType,
    BackendResourceType,
    BackendUploadType,
    CountResult,
    LoginAttempt,
    LoginResult,
    Match,
    NotFoundError,
    OKMessage,
    UploadStage,
    UploadStatus,
)
from matchbox.common.exceptions import (
    MatchboxDeletionNotConfirmed,
    MatchboxResolutionNotFoundError,
    MatchboxServerFileError,
)
from matchbox.common.graph import ResolutionGraph, ResolutionName, SourceResolutionName
from matchbox.server.api.dependencies import (
    BackendDependency,
    ParquetResponse,
    SettingsDependency,
    UploadTrackerDependency,
    authorisation_dependencies,
    lifespan,
)
from matchbox.server.api.routers import eval, resolution
from matchbox.server.uploads import process_upload, process_upload_celery, table_to_s3

app = FastAPI(
    title="matchbox API",
    version=version("matchbox_db"),
    lifespan=lifespan,
)
app.include_router(resolution.router)
app.include_router(eval.router)


@app.exception_handler(StarletteHTTPException)
async def http_exception_handler(request, exc):
    """Overwrite the default JSON schema for an `HTTPException`."""
    return JSONResponse(
        content=jsonable_encoder(exc.detail), status_code=exc.status_code
    )


@app.middleware("http")
async def add_security_headers(request: Request, call_next):
    """Improve security by adding headers to all responses."""
    response: Response = await call_next(request)
    response.headers["Cache-control"] = "no-store, no-cache"
    response.headers["Content-Security-Policy"] = (
        "default-src 'none'; frame-ancestors 'none'; form-action 'none'; sandbox"
    )
    response.headers["Strict-Transport-Security"] = (
        "max-age=31536000; includeSubDomains"
    )
    response.headers["X-Frame-Options"] = "DENY"
    response.headers["X-Content-Type-Options"] = "nosniff"
    return response


@app.get("/health")
async def healthcheck() -> OKMessage:
    """Perform a health check and return the status."""
    return OKMessage()


@app.post(
    "/login",
)
def login(
    backend: BackendDependency,
    credentials: LoginAttempt,
) -> LoginResult:
    """Receives a user name and returns a user ID."""
    return LoginResult(user_id=backend.login(credentials.user_name))


@app.post(
    "/upload/{upload_id}",
    responses={
        400: {"model": UploadStatus, **UploadStatus.status_400_examples()},
    },
    status_code=status.HTTP_202_ACCEPTED,
    dependencies=[Depends(authorisation_dependencies)],
)
def upload_file(
    backend: BackendDependency,
    upload_tracker: UploadTrackerDependency,
    settings: SettingsDependency,
    background_tasks: BackgroundTasks,
    upload_id: str,
    file: UploadFile,
) -> UploadStatus:
    """Upload and process a file based on metadata type.

    The file is uploaded to S3 and then processed in a background task.
    Status can be checked using the /upload/{upload_id}/status endpoint.

    Raises HTTP 400 if:

    * Upload ID not found or expired (entries expire after 30 minutes of inactivity)
    * Upload is already being processed
    * Uploaded data doesn't match the metadata schema
    """
    # Get and validate cache entry
    upload_entry = upload_tracker.get(upload_id=upload_id)
    if not upload_entry:
        raise HTTPException(
            status_code=400,
            detail=UploadStatus(
                id=upload_id,
                update_timestamp=datetime.now(),
                stage="unknown",
                details=(
                    "Upload ID not found or expired. Entries expire after 30 minutes "
                    "of inactivity, including failed processes."
                ),
            ).model_dump(),
        )

    # Ensure tracker is expecting file
    if upload_entry.status.stage != UploadStage.AWAITING_UPLOAD:
        raise HTTPException(
            status_code=400,
            detail=upload_entry.status.model_dump(),
        )

    # Upload to S3
    client = backend.settings.datastore.get_client()
    bucket = backend.settings.datastore.cache_bucket_name
    key = f"{upload_id}.parquet"

    try:
        table_to_s3(
            client=client,
            bucket=bucket,
            key=key,
            file=file,
            expected_schema=upload_entry.status.entity.schema,
        )
    except MatchboxServerFileError as e:
        upload_tracker.update(upload_id, UploadStage.FAILED, details=str(e))
        updated_entry = upload_tracker.get(upload_id)
        raise HTTPException(
            status_code=400,
            detail=updated_entry.status.model_dump(),
        ) from e

    upload_tracker.update(upload_id, UploadStage.QUEUED)

    # Start background processing
    match settings.task_runner:
        case "api":
            background_tasks.add_task(
                process_upload,
                backend=backend,
                tracker=upload_tracker,
                s3_client=client,
                upload_type=upload_entry.status.entity,
                resolution_name=upload_entry.metadata.name,
                upload_id=upload_id,
                bucket=bucket,
                filename=key,
            )
        case "celery":
            process_upload_celery.delay(
                upload_type=upload_entry.status.entity,
                resolution_name=upload_entry.metadata.name,
                upload_id=upload_id,
                bucket=bucket,
                filename=key,
            )
        case _:
            raise RuntimeError("Unsupported task runner.")

    source_upload = upload_tracker.get(upload_id)

    # Check for error in async task
    if source_upload.status.stage == UploadStage.FAILED:
        raise HTTPException(
            status_code=400,
            detail=source_upload.status.model_dump(),
        )
    else:
        return source_upload.status


@app.get(
    "/upload/{upload_id}/status",
    responses={
        400: {"model": UploadStatus, **UploadStatus.status_400_examples()},
    },
    status_code=status.HTTP_200_OK,
)
def get_upload_status(
    upload_tracker: UploadTrackerDependency,
    upload_id: str,
) -> UploadStatus:
    """Get the status of an upload process.

    Returns the current status of the upload and processing task.

    Raises HTTP 400 if:
    * Upload ID not found or expired (entries expire after 30 minutes of inactivity)
    """
    source_upload = upload_tracker.get(upload_id=upload_id)
    if not source_upload:
        raise HTTPException(
            status_code=400,
            detail=UploadStatus(
                id=upload_id,
                stage="unknown",
                update_timestamp=datetime.now(),
                details=(
                    "Upload ID not found or expired. Entries expire after 30 minutes "
                    "of inactivity, including failed processes."
                ),
                entity=BackendUploadType.INDEX,
            ).model_dump(),
        )

    return source_upload.status


# Retrieval


@app.get(
    "/query",
    responses={404: {"model": NotFoundError}},
)
def query(
    backend: BackendDependency,
    source: SourceResolutionName,
    return_leaf_id: bool,
    resolution: ResolutionName | None = None,
    threshold: int | None = None,
    limit: int | None = None,
) -> ParquetResponse:
    """Query Matchbox for matches based on a source resolution name."""
    try:
        res = backend.query(
            source=source,
            resolution=resolution,
            threshold=threshold,
            return_leaf_id=return_leaf_id,
            limit=limit,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e

    buffer = table_to_buffer(res)
    return ParquetResponse(buffer.getvalue())


@app.get(
    "/match",
    responses={404: {"model": NotFoundError}},
)
def match(
    backend: BackendDependency,
    targets: Annotated[list[SourceResolutionName], Query()],
    source: SourceResolutionName,
    key: str,
    resolution: ResolutionName,
    threshold: int | None = None,
) -> list[Match]:
    """Match a source key against a list of target source resolutions."""
    try:
        res = backend.match(
            key=key,
            source=source,
            targets=targets,
            resolution=resolution,
            threshold=threshold,
        )
    except MatchboxResolutionNotFoundError as e:
        raise HTTPException(
            status_code=404,
            detail=NotFoundError(
                details=str(e), entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        ) from e

    return res


# Admin


@app.get("/report/resolutions")
def get_resolutions(backend: BackendDependency) -> ResolutionGraph:
    """Get the resolution graph."""
    return backend.get_resolution_graph()


@app.get("/database/count")
def count_backend_items(
    backend: BackendDependency,
    entity: BackendCountableType | None = None,
) -> CountResult:
    """Count the number of various entities in the backend."""

    def get_count(e: BackendCountableType) -> int:
        return getattr(backend, str(e)).count()

    if entity is not None:
        return CountResult(entities={str(entity): get_count(entity)})
    else:
        res = {str(e): get_count(e) for e in BackendCountableType}
        return CountResult(entities=res)


@app.delete(
    "/database",
    responses={409: {"model": str}},
    dependencies=[Depends(authorisation_dependencies)],
)
def clear_database(
    backend: BackendDependency,
    certain: Annotated[
        bool,
        Query(
            description=(
                "Confirm deletion of all data in the database whilst retaining tables"
            )
        ),
    ] = False,
) -> OKMessage:
    """Delete all data from the backend whilst retaining tables."""
    try:
        backend.clear(certain=certain)
        return OKMessage()
    except MatchboxDeletionNotConfirmed as e:
        raise HTTPException(
            status_code=409,
            detail=str(e),
        ) from e



"""Matchbox PostgreSQL database connection."""

from collections.abc import Generator
from contextlib import contextmanager
from pathlib import Path
from typing import Any

from adbc_driver_postgresql import dbapi as adbc_dbapi
from alembic import command
from alembic.config import Config
from pydantic import BaseModel, Field
from sqlalchemy import (
    URL,
    Engine,
    MetaData,
    Table,
    create_engine,
    inspect,
    text,
)
from sqlalchemy.orm import Session, declarative_base, sessionmaker
from sqlalchemy.pool import PoolProxiedConnection, QueuePool

from matchbox.common.logging import logger
from matchbox.server.base import MatchboxBackends, MatchboxServerSettings


class MatchboxPostgresCoreSettings(BaseModel):
    """PostgreSQL-specific settings for Matchbox."""

    host: str
    port: int
    user: str
    password: str
    database: str
    db_schema: str
    alembic_config: Path = Field(
        default=Path("src/matchbox/server/postgresql/alembic.ini")
    )

    def get_alembic_config(self) -> Config:
        """Get the Alembic config."""
        config = Config(self.alembic_config)
        db_url = URL.create(
            "postgresql+psycopg",
            username=self.user,
            password=self.password,
            host=self.host,
            port=self.port,
            database=self.database,
        )
        config.set_main_option(
            "sqlalchemy.url", db_url.render_as_string(hide_password=False)
        )
        return config


class MatchboxPostgresSettings(MatchboxServerSettings):
    """Settings for the Matchbox PostgreSQL backend.

    Inherits the core settings and adds the PostgreSQL-specific settings.
    """

    backend_type: MatchboxBackends = MatchboxBackends.POSTGRES

    postgres: MatchboxPostgresCoreSettings = Field(
        default_factory=MatchboxPostgresCoreSettings
    )


class MatchboxDatabase:
    """Matchbox PostgreSQL database connection."""

    def __init__(self, settings: MatchboxPostgresSettings):
        """Initialise the database connection."""
        self.settings = settings
        self._engine: Engine | None = None
        self._SessionLocal: sessionmaker | None = None
        self._adbc_pool: QueuePool | None = None
        self._source_adbc_connection: adbc_dbapi.Connection | None = None
        self.MatchboxBase = declarative_base(
            metadata=MetaData(schema=settings.postgres.db_schema)
        )
        self.alembic_config = settings.postgres.get_alembic_config()

    def connection_string(self, driver: bool = True) -> str:
        """Get the connection string for PostgreSQL."""
        driver_string = ""
        if driver:
            driver_string = "+psycopg"
        return (
            f"postgresql{driver_string}://{self.settings.postgres.user}:{self.settings.postgres.password}"
            f"@{self.settings.postgres.host}:{self.settings.postgres.port}/"
            f"{self.settings.postgres.database}"
        )

    def _connect(self):
        """Connect to the database."""
        self._engine = create_engine(
            url=self.connection_string(), logging_name="matchbox.engine", echo=False
        )
        self._SessionLocal = sessionmaker(
            autocommit=False, autoflush=False, bind=self._engine
        )

    def _disconnect(self):
        if self._engine:
            self._engine.dispose()
            self._engine = None
            self._SessionLocal = None

    def _connect_adbc(self) -> None:
        self._source_adbc_connection = adbc_dbapi.connect(
            self.connection_string(driver=False)
        )
        self._adbc_pool = QueuePool(
            self._source_adbc_connection.adbc_clone,
        )

    def _disconnect_adbc(self) -> None:
        if self._adbc_pool:
            self._adbc_pool.dispose()
            self._adbc_pool = None

            self._source_adbc_connection.close()
            self._source_adbc_connection = None

    def _reset_connections(self) -> None:
        """Dispose and re-initialise SQLAlchemy and ADBC connection managers."""
        self._disconnect()
        self._connect()

        self._disconnect_adbc()
        self._connect_adbc()

    def get_engine(self) -> Engine:
        """Get the database engine."""
        if not self._engine:
            self._connect()
        return self._engine

    def get_session(self) -> Session:
        """Get a new session."""
        if not self._SessionLocal:
            self._connect()
        return self._SessionLocal()

    @contextmanager
    def get_adbc_connection(self) -> Generator[PoolProxiedConnection, Any, Any]:
        """Get a new ADBC connection wrapped by a SQLAlchemy pool proxy.

        The connection must be used within a context manager.
        """
        if not self._adbc_pool:
            self._connect_adbc()

        conn = self._adbc_pool.connect()
        try:
            yield conn
        finally:
            conn.close()

    def run_migrations(self):
        """Create the database and all tables expected in the schema."""
        alembic_version = self._look_for_alembic_version()
        engine = self.get_engine()
        logger.info("Determinded alembic in use so upgrading to head")
        if alembic_version is not None:
            command.upgrade(self.alembic_config, "head")
        else:
            logger.info(
                "Determinded alembic not in use so dropping schema if it "
                "exists prior to upgrading to head. "
            )
            with engine.connect() as conn:
                conn.execute(text("DROP SCHEMA IF EXISTS mb CASCADE;"))
                conn.commit()
            command.upgrade(self.alembic_config, "head")

    def clear_database(self):
        """Delete all rows in every table in the database schema.

        - TRUNCATE tables that are part of the core ORM (preserves structure)
        - DROP tables that are not in the ORM (removes temporary/test tables)
        """
        schema_name = self.MatchboxBase.metadata.schema
        engine = self.get_engine()  # Get the engine

        with self.get_session() as session:
            try:
                # Get all tables that actually exist in the database
                discovered_metadata = MetaData(schema=schema_name)
                discovered_metadata.reflect(
                    bind=engine, schema=schema_name
                )  # Use engine here

                # Get table names from the core ORM
                orm_table_names = {
                    table.name for table in self.MatchboxBase.metadata.tables.values()
                }

                # Categorise discovered tables
                tables_to_truncate: list[Table] = []
                tables_to_drop: list[Table] = []

                for table in discovered_metadata.tables.values():
                    if table.name == "alembic_version":
                        # Alembic version is in public but just in case
                        continue
                    elif table.name in orm_table_names:
                        tables_to_truncate.append(table)
                    else:
                        tables_to_drop.append(table)

                # TRUNCATE core ORM tables (preserves schema structure)
                if tables_to_truncate:
                    for table in tables_to_truncate:
                        session.execute(
                            text(f"TRUNCATE TABLE {table.fullname} CASCADE")
                        )

                # DROP temporary/test tables (removes them completely)
                if tables_to_drop:
                    for table in tables_to_drop:
                        session.execute(text(f"DROP TABLE {table.fullname} CASCADE"))

                session.commit()

            except Exception as e:
                logger.error(f"Error clearing database: {e}")
                session.rollback()
                raise

    def drop_database(self):
        """Drop all tables in the database schema and re-recreate them."""
        command.downgrade(self.alembic_config, "base")
        command.upgrade(self.alembic_config, "head")

    def _look_for_alembic_version(self) -> bool:
        engine = self.get_engine()
        inspector = inspect(engine)
        alembic_version_table = "alembic_version" in inspector.get_table_names(
            schema="public"
        )
        if alembic_version_table:
            with engine.connect() as conn:
                result = conn.execute(
                    text("SELECT version_num FROM public.alembic_version;")
                )
                alembic_version = result.scalar()
        else:
            alembic_version = None
        return alembic_version

    @property
    def sorted_tables(self) -> list[Table]:
        """Return a list of SQLAlchemy tables in order of creation."""
        return self.MatchboxBase.metadata.sorted_tables


# Global database instance -- everything should use this

MBDB = MatchboxDatabase(MatchboxPostgresSettings())



import json
from datetime import datetime

import pyarrow as pa
import pytest
from httpx import Response
from respx.router import MockRouter
from sqlalchemy import Engine

from matchbox.client.dags import DAG
from matchbox.client.models import Model, add_model_class
from matchbox.client.models.linkers.base import LinkerSettings
from matchbox.client.queries import Query
from matchbox.client.results import Results
from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import (
    BackendResourceType,
    BackendUploadType,
    CRUDOperation,
    ModelConfig,
    ModelType,
    NotFoundError,
    ResolutionOperationStatus,
    UploadStage,
    UploadStatus,
)
from matchbox.common.exceptions import (
    MatchboxDeletionNotConfirmed,
    MatchboxServerFileError,
)
from matchbox.common.factories.models import MockLinker, model_factory
from matchbox.common.factories.sources import source_factory


def test_init_and_run_model(sqlite_warehouse: Engine, matchbox_api: MockRouter):
    """Test that model can be initialised and run correctly."""
    # Register "custom" model
    add_model_class(MockLinker)

    # Mock API
    foo = source_factory(engine=sqlite_warehouse).write_to_location()
    bar = source_factory(engine=sqlite_warehouse).write_to_location()

    foo_leafy_data = foo.data.append_column(
        "leaf_id", pa.array(range(len(foo.data)), type=pa.int64())
    )
    bar_leafy_data = foo.data.append_column(
        "leaf_id",
        pa.array(
            range(len(foo_leafy_data), len(foo_leafy_data) + len(bar.data)),
            type=pa.int64(),
        ),
    )

    # Mock API
    matchbox_api.get("/query").mock(
        side_effect=[
            # First query
            Response(200, content=table_to_buffer(foo.data).read()),
            Response(200, content=table_to_buffer(bar.data).read()),
            # Second query (for validation)
            Response(200, content=table_to_buffer(foo_leafy_data).read()),
            Response(200, content=table_to_buffer(bar_leafy_data).read()),
        ]
    )
    dag = DAG("collection")
    foo_query = Query(foo.source, dag=dag)
    bar_query = Query(bar.source, dag=dag)

    model = Model(
        dag=dag,
        name="name",
        description="description",
        model_class=MockLinker,
        model_settings=LinkerSettings(),
        left_query=foo_query,
        right_query=bar_query,
    )

    assert model.config == ModelConfig(
        type=ModelType.LINKER,
        model_class="MockLinker",
        model_settings=json.dumps({"left_id": "l.field", "right_id": "r.field"}),
        left_query=foo_query.config,
        right_query=bar_query.config,
    )

    model.run()
    # This won't actually run anything new
    with pytest.warns(match="already run"):
        model.run(for_validation=True)
    assert model.results.left_root_leaf is None
    assert model.results.right_root_leaf is None

    model.run(for_validation=True, full_rerun=True)
    assert model.results.left_root_leaf is not None
    assert model.results.right_root_leaf is not None


def test_model_sync(matchbox_api: MockRouter):
    """Test syncing a model, its truth and results."""
    # Create test model using factory
    testkit = model_factory(model_type="linker")

    # Mock endpoints
    get_route = matchbox_api.get(f"/resolutions/{testkit.model.name}").mock(
        return_value=Response(
            404,
            json=NotFoundError(
                details="Model not found", entity=BackendResourceType.RESOLUTION
            ).model_dump(),
        )
    )
    insert_config_route = matchbox_api.post("/resolutions").mock(
        return_value=Response(
            201,
            json=ResolutionOperationStatus(
                success=True,
                name=testkit.model.name,
                operation=CRUDOperation.CREATE,
            ).model_dump(),
        )
    )

    set_truth_route = matchbox_api.patch(
        f"/resolutions/{testkit.model.name}/truth"
    ).mock(
        return_value=Response(
            200,
            json=ResolutionOperationStatus(
                success=True,
                name=testkit.model.name,
                operation=CRUDOperation.UPDATE,
            ).model_dump(),
        )
    )

    insert_results_route = matchbox_api.post(
        f"/resolutions/{testkit.model.name}/data"
    ).mock(
        return_value=Response(
            202,
            content=UploadStatus(
                id="test-upload-id",
                stage=UploadStage.AWAITING_UPLOAD,
                update_timestamp=datetime.now(),
                entity=BackendUploadType.RESULTS,
            ).model_dump_json(),
        )
    )

    upload_route = matchbox_api.post("/upload/test-upload-id").mock(
        return_value=Response(
            202,
            content=UploadStatus(
                id="test-upload-id",
                stage=UploadStage.PROCESSING,
                update_timestamp=datetime.now(),
                entity=BackendUploadType.RESULTS,
            ).model_dump_json(),
        )
    )

    status_route = matchbox_api.get("/upload/test-upload-id/status").mock(
        return_value=Response(
            200,
            content=UploadStatus(
                id="test-upload-id",
                stage=UploadStage.COMPLETE,
                update_timestamp=datetime.now(),
                entity=BackendUploadType.RESULTS,
            ).model_dump_json(),
        )
    )

    # Call sync
    testkit.model.sync()

    # Verify the API call
    assert get_route.called
    assert insert_config_route.called
    assert (
        insert_config_route.calls.last.request.content.decode()
        == testkit.model.to_resolution().model_dump_json()
    )
    assert set_truth_route.called
    assert float(set_truth_route.calls.last.request.read()) == 100
    assert not insert_results_route.called

    # Set results
    test_results = Results(probabilities=testkit.probabilities)

    testkit.model.results = test_results
    testkit.model.sync()

    # Verify API calls
    assert insert_results_route.called
    assert upload_route.called
    assert status_route.called
    assert (
        b"PAR1" in upload_route.calls.last.request.content
    )  # Check for parquet file signature

    # Mock the upload endpoint with a failure
    matchbox_api.post("/upload/test-upload-id").mock(
        return_value=Response(
            400,
            content=UploadStatus(
                id="test-upload-id",
                stage=UploadStage.FAILED,
                update_timestamp=datetime.now(),
                entity=BackendUploadType.RESULTS,
                details="Invalid data format",
            ).model_dump_json(),
        )
    )

    with pytest.raises(MatchboxServerFileError, match="Invalid data format"):
        testkit.model.sync()

    # Mock earlier endpoint generating a name clash
    source = source_factory().source
    matchbox_api.get(f"/resolutions/{testkit.model.name}").mock(
        return_value=Response(200, json=source.to_resolution().model_dump())
    )

    with pytest.raises(ValueError, match="existing resolution"):
        testkit.model.sync()


def test_truth_getter():
    """Test getting model truth threshold from config."""
    # Create testkit with specific truth value
    testkit = model_factory(model_type="linker")
    # Update the model to have a truth value
    testkit.model._truth = 90  # Integer truth value (90 = 0.9 as float)

    # Get truth as float
    truth = testkit.model.truth

    # Verify it returns the correct value converted to float
    assert truth == 0.9


def test_truth_setter_validation_error():
    """Test setting invalid truth values."""
    testkit = model_factory(model_type="linker")

    # Attempt to set an invalid truth value using the validated setter
    with pytest.raises(ValueError):
        testkit.model.truth = 1.5


def test_delete_resolution(matchbox_api: MockRouter):
    """Test successfully deleting a resolution."""
    # Create test model using factory
    testkit = model_factory()

    # Mock the DELETE endpoint with success response
    route = matchbox_api.delete(
        f"/resolutions/{testkit.model.name}", params={"certain": True}
    ).mock(
        return_value=Response(
            200,
            json=ResolutionOperationStatus(
                success=True,
                name=testkit.model.name,
                operation=CRUDOperation.DELETE,
            ).model_dump(),
        )
    )

    # Delete the model
    response = testkit.model.delete(certain=True)

    # Verify the response and API call
    assert response
    assert route.called
    assert route.calls.last.request.url.params["certain"] == "true"


def test_delete_resolution_needs_confirmation(matchbox_api: MockRouter):
    """Test attempting to delete a resolution without confirmation returns 409."""
    # Create test model using factory
    testkit = model_factory()

    # Mock the DELETE endpoint with 409 confirmation required response
    error_details = "Cannot delete model with dependent models: dedupe1, dedupe2"
    route = matchbox_api.delete(f"/resolutions/{testkit.model.name}").mock(
        return_value=Response(
            409,
            json=ResolutionOperationStatus(
                success=False,
                name=testkit.model.name,
                operation=CRUDOperation.DELETE,
                details=error_details,
            ).model_dump(),
        )
    )

    # Attempt to delete without certain=True
    with pytest.raises(MatchboxDeletionNotConfirmed):
        testkit.model.delete()

    # Verify the response and API call
    assert route.called
    assert route.calls.last.request.url.params["certain"] == "false"



from time import sleep
from typing import TYPE_CHECKING, Any
from unittest.mock import Mock

import pytest
from botocore.exceptions import ClientError
from fastapi.testclient import TestClient

from matchbox.common.arrow import table_to_buffer
from matchbox.common.dtos import (
    BackendResourceType,
    CRUDOperation,
    NotFoundError,
    Resolution,
    ResolutionOperationStatus,
    UploadStage,
)
from matchbox.common.exceptions import (
    MatchboxDeletionNotConfirmed,
    MatchboxResolutionNotFoundError,
)
from matchbox.common.factories.models import model_factory
from matchbox.common.factories.sources import source_factory

if TYPE_CHECKING:
    from mypy_boto3_s3.client import S3Client
else:
    S3Client = Any


def test_get_source(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    source_testkit = source_factory(name="foo")
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution = Mock(
        return_value=source_testkit.source.to_resolution(),  # Second call (SOURCE)
    )

    response = test_client.get("/resolutions/foo")
    assert response.status_code == 200
    assert response.json()["name"] == "foo"
    assert response.json()["resolution_type"] == "source"


def test_get_model(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    testkit = model_factory(name="test_model", description="test description")
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution = Mock(return_value=testkit.model.to_resolution())

    response = test_client.get("/resolutions/test_model")

    assert response.status_code == 200
    assert response.json()["name"] == testkit.model.name
    assert response.json()["description"] == testkit.model.description


def test_get_resolution_404(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution = Mock(side_effect=MatchboxResolutionNotFoundError())

    response = test_client.get("/resolutions/nonexistent")

    assert response.status_code == 404
    assert response.json()["entity"] == BackendResourceType.RESOLUTION


def test_insert_model(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    testkit = model_factory(name="test_model")
    test_client, mock_backend, _ = api_client_and_mocks

    response = test_client.post(
        "/resolutions", json=testkit.model.to_resolution().model_dump()
    )

    assert response.status_code == 201
    assert (
        response.json()
        == ResolutionOperationStatus(
            success=True,
            name="test_model",
            operation=CRUDOperation.CREATE,
            details=None,
        ).model_dump()
    )

    mock_backend.insert_resolution.assert_called_once_with(
        resolution=testkit.model.to_resolution()
    )


def test_insert_model_error(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.insert_resolution = Mock(side_effect=Exception("Test error"))
    testkit = model_factory()

    response = test_client.post(
        "/resolutions", json=testkit.model.to_resolution().model_dump()
    )

    assert response.status_code == 500
    assert response.json()["success"] is False
    assert response.json()["details"] == "Test error"


@pytest.mark.parametrize("model_type", ["deduper", "linker"])
def test_complete_model_upload_process(
    s3: S3Client,
    model_type: str,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    """Test the complete upload process for models from creation through processing."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.settings.datastore.get_client.return_value = s3
    mock_backend.settings.datastore.cache_bucket_name = "test-bucket"
    mock_backend.insert_model_data = Mock(return_value=None)

    # Create test bucket
    s3.create_bucket(
        Bucket="test-bucket",
        CreateBucketConfiguration={"LocationConstraint": "eu-west-2"},
    )

    # Create test data with specified model type
    testkit = model_factory(model_type=model_type)

    # Set up the mock to return the actual model metadata and data
    mock_backend.get_resolution = Mock(return_value=testkit.model.to_resolution())
    mock_backend.get_model_data = Mock(return_value=testkit.probabilities)

    # Step 1: Create model
    response = test_client.post(
        "/resolutions", json=testkit.model.to_resolution().model_dump()
    )
    assert response.status_code == 201
    assert response.json()["success"] is True
    assert response.json()["name"] == testkit.model.name

    # Step 2: Initialize results upload
    response = test_client.post(f"/resolutions/{testkit.model.name}/data")
    assert response.status_code == 202
    upload_id = response.json()["id"]
    assert response.json()["stage"] == UploadStage.AWAITING_UPLOAD

    # Step 3: Upload results file with real background tasks
    response = test_client.post(
        f"/upload/{upload_id}",
        files={
            "file": (
                "results.parquet",
                table_to_buffer(testkit.probabilities),
                "application/octet-stream",
            ),
        },
    )
    assert response.status_code == 202
    assert response.json()["stage"] == UploadStage.QUEUED

    # Step 4: Poll status until complete or timeout
    max_attempts = 10
    current_attempt = 0
    stage = None

    while current_attempt < max_attempts:
        response = test_client.get(f"/upload/{upload_id}/status")
        assert response.status_code == 200

        stage = response.json()["stage"]
        if stage == UploadStage.COMPLETE:
            break
        elif stage == UploadStage.FAILED:
            pytest.fail(f"Upload failed: {response.json().get('details')}")
        elif stage in [UploadStage.PROCESSING, UploadStage.QUEUED]:
            sleep(0.1)  # Small delay between polls
        else:
            pytest.fail(f"Unexpected stage: {stage}")

        current_attempt += 1

    assert current_attempt < max_attempts, (
        "Timed out waiting for processing to complete"
    )
    assert stage == UploadStage.COMPLETE
    assert response.status_code == 200

    # Step 5: Verify results were stored correctly
    mock_backend.insert_model_data.assert_called_once()
    call_args = mock_backend.insert_model_data.call_args
    assert (
        call_args[1]["name"] == testkit.model.name
    )  # Check model resolution name matches
    assert call_args[1]["results"].equals(
        testkit.probabilities
    )  # Check results data matches

    # Step 6: Verify we can retrieve the results
    response = test_client.get(f"/resolutions/{testkit.model.name}/data")
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/octet-stream"

    # Verify the model truth can be set and retrieved
    truth_value = 85
    mock_backend.get_model_truth = Mock(return_value=truth_value)

    response = test_client.patch(
        f"/resolutions/{testkit.model.name}/truth",
        json=truth_value,
    )
    assert response.status_code == 200

    response = test_client.get(f"/resolutions/{testkit.model.name}/truth")
    assert response.status_code == 200
    assert response.json() == truth_value

    # Verify file is deleted from S3 after processing
    with pytest.raises(ClientError):
        s3.head_object(Bucket="test-bucket", Key=f"{upload_id}.parquet")


def test_set_results(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    testkit = model_factory()
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution = Mock(return_value=testkit.model.to_resolution())

    response = test_client.post(f"/resolutions/{testkit.model.name}/data")

    assert response.status_code == 202
    assert response.json()["stage"] == UploadStage.AWAITING_UPLOAD


def test_set_results_model_not_found(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    """Test setting results for a non-existent model."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_resolution = Mock(side_effect=MatchboxResolutionNotFoundError())

    response = test_client.post("/resolutions/nonexistent-model/data")

    assert response.status_code == 404
    assert response.json()["entity"] == BackendResourceType.RESOLUTION


def test_get_results(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    testkit = model_factory()
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_model_data = Mock(return_value=testkit.probabilities)

    response = test_client.get(f"/resolutions/{testkit.model.name}/data")

    assert response.status_code == 200
    assert response.headers["content-type"] == "application/octet-stream"


def test_set_truth(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    testkit = model_factory()
    test_client, mock_backend, _ = api_client_and_mocks

    response = test_client.patch(f"/resolutions/{testkit.model.name}/truth", json=95)

    assert response.status_code == 200
    assert response.json()["success"] is True
    mock_backend.set_model_truth.assert_called_once_with(
        name=testkit.model.name, truth=95
    )


def test_set_truth_invalid_value(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test setting an invalid truth value (outside 0-1 range)."""
    testkit = model_factory()
    test_client, _, _ = api_client_and_mocks

    # Test value > 1
    response = test_client.patch(f"/resolutions/{testkit.model.name}/truth", json=150)
    assert response.status_code == 422

    # Test value < 0
    response = test_client.patch(f"/resolutions/{testkit.model.name}/truth", json=-50)
    assert response.status_code == 422


def test_get_truth(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    testkit = model_factory()
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_model_truth = Mock(return_value=95)

    response = test_client.get(f"/resolutions/{testkit.model.name}/truth")

    assert response.status_code == 200
    assert response.json() == 95


@pytest.mark.parametrize(
    "endpoint",
    ["data", "truth"],
)
def test_model_get_endpoints_404(
    endpoint: str,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test 404 responses for model GET endpoints when model doesn't exist."""
    test_client, mock_backend, _ = api_client_and_mocks
    # Map endpoint to actual backend method name
    method_name = "get_model_data" if endpoint == "data" else f"get_model_{endpoint}"
    mock_method = getattr(mock_backend, method_name)
    mock_method.side_effect = MatchboxResolutionNotFoundError()

    response = test_client.get(f"/resolutions/nonexistent-model/{endpoint}")

    assert response.status_code == 404
    error = NotFoundError.model_validate(response.json())
    assert error.entity == BackendResourceType.RESOLUTION


@pytest.mark.parametrize(
    ("endpoint", "payload"),
    [
        ("truth", 95),
    ],
)
def test_model_patch_endpoints_404(
    endpoint: str,
    payload: float | list[dict[str, Any]],
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test 404 responses for model PATCH endpoints when model doesn't exist."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_method = getattr(mock_backend, f"set_model_{endpoint}")
    mock_method.side_effect = MatchboxResolutionNotFoundError()

    response = test_client.patch(
        f"/resolutions/nonexistent-model/{endpoint}", json=payload
    )

    assert response.status_code == 404
    error = NotFoundError.model_validate(response.json())
    assert error.entity == BackendResourceType.RESOLUTION


def test_delete_resolution(api_client_and_mocks: tuple[TestClient, Mock, Mock]):
    """Test deletion of a resolution."""
    testkit = model_factory()
    test_client, _, _ = api_client_and_mocks
    response = test_client.delete(
        f"/resolutions/{testkit.model.name}",
        params={"certain": True},
    )

    assert response.status_code == 200
    assert (
        response.json()
        == ResolutionOperationStatus(
            success=True,
            name=testkit.model.name,
            operation=CRUDOperation.DELETE,
            details=None,
        ).model_dump()
    )


def test_delete_resolution_needs_confirmation(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    """Test deletion of a model that requires confirmation."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.delete_resolution = Mock(
        side_effect=MatchboxDeletionNotConfirmed(children=["dedupe1", "dedupe2"])
    )

    testkit = model_factory()
    response = test_client.delete(f"/resolutions/{testkit.model.name}")

    assert response.status_code == 409
    assert response.json()["success"] is False
    message = response.json()["details"]
    assert "dedupe1" in message and "dedupe2" in message


@pytest.mark.parametrize("certain", [True, False])
def test_delete_resolution_404(
    certain: bool,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
) -> None:
    """Test 404 response when trying to delete a non-existent resolution."""
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.delete_resolution.side_effect = MatchboxResolutionNotFoundError()

    response = test_client.delete(
        "/resolutions/nonexistent-model", params={"certain": certain}
    )

    assert response.status_code == 404
    error = NotFoundError.model_validate(response.json())
    assert error.entity == BackendResourceType.RESOLUTION


def test_get_resolution_sources(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    source = source_factory().source.to_resolution()
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_leaf_source_resolutions = Mock(return_value=[source])

    response = test_client.get("/resolutions/foo/sources")
    assert response.status_code == 200
    for s in response.json():
        assert Resolution.model_validate(s)


def test_get_resolution_sources_404(
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.get_leaf_source_resolutions = Mock(
        side_effect=MatchboxResolutionNotFoundError
    )

    response = test_client.get("/resolutions/foo/sources")
    assert response.status_code == 404
    assert response.json()["entity"] == BackendResourceType.RESOLUTION


def test_complete_source_upload_process(
    s3: S3Client,
    api_client_and_mocks: tuple[TestClient, Mock, Mock],
):
    """Test the complete upload process from source creation through processing."""
    # Create test data
    source_testkit = source_factory()

    # Setup the backend
    test_client, mock_backend, _ = api_client_and_mocks
    mock_backend.settings.datastore.get_client.return_value = s3
    mock_backend.settings.datastore.cache_bucket_name = "test-bucket"

    mock_backend.get_resolution = Mock(
        return_value=source_testkit.source.to_resolution()
    )
    mock_backend.insert_resolution = Mock(return_value=None)
    mock_backend.insert_source_data = Mock(return_value=None)

    # Create test bucket
    s3.create_bucket(
        Bucket="test-bucket",
        CreateBucketConfiguration={"LocationConstraint": "eu-west-2"},
    )

    # Step 1: Add source
    response = test_client.post(
        "/resolutions",
        json=source_testkit.source.to_resolution().model_dump(mode="json"),
    )
    assert response.status_code == 201
    status = ResolutionOperationStatus.model_validate(response.json())
    assert response.status_code == 201, response.json()
    assert status.name == source_testkit.name

    # Assert backend given the config but not yet the data
    mock_backend.insert_resolution.assert_called_once_with(
        resolution=source_testkit.source.to_resolution()
    )
    mock_backend.insert_source_data.assert_not_called()

    response = test_client.post(f"/resolutions/{source_testkit.name}/data")
    upload_id = response.json()["id"]
    assert response.status_code == 202
    assert response.json()["stage"] == UploadStage.AWAITING_UPLOAD

    # Step 2: Upload file with real background tasks
    response = test_client.post(
        f"/upload/{upload_id}",
        files={
            "file": (
                "hashes.parquet",
                table_to_buffer(source_testkit.data_hashes),
                "application/octet-stream",
            ),
        },
    )
    assert response.status_code == 202
    assert response.json()["stage"] == UploadStage.QUEUED

    # Step 3: Poll status until complete or timeout
    max_attempts = 10
    current_attempt = 0
    while current_attempt < max_attempts:
        response = test_client.get(f"/upload/{upload_id}/status")
        assert response.status_code == 200

        stage = response.json()["stage"]
        if stage == UploadStage.COMPLETE:
            break
        elif stage == UploadStage.FAILED:
            pytest.fail(f"Upload failed: {response.json().get('details')}")
        elif stage in [UploadStage.PROCESSING, UploadStage.QUEUED]:
            sleep(0.1)  # Small delay between polls
        else:
            pytest.fail(f"Unexpected stage: {stage}")

        current_attempt += 1

    assert current_attempt < max_attempts, (
        "Timed out waiting for processing to complete"
    )
    assert stage == UploadStage.COMPLETE
    assert response.status_code == 200

    # Verify backend methods were called with correct arguments
    mock_backend.insert_source_data.assert_called_once()

    # Check resolution matches
    call_args = mock_backend.insert_source_data.call_args
    assert call_args[1]["data_hashes"].equals(source_testkit.data_hashes)  # Check data

    # Verify file is deleted from S3 after processing
    with pytest.raises(ClientError):
        s3.head_object(Bucket="test-bucket", Key=f"{upload_id}.parquet")



"""Interface for submitting evaluation data."""

import polars as pl
import streamlit as st
from sqlalchemy import Engine, create_engine

from matchbox.client import _handler
from matchbox.client._settings import settings
from matchbox.client.dags import DAG
from matchbox.client.eval import get_samples
from matchbox.common.eval import Judgement
from matchbox.common.exceptions import MatchboxClientSettingsException


@st.cache_resource
def init_engine() -> Engine:
    """Create a single SQLAlchemy engine."""
    engine = create_engine(settings.default_warehouse)
    return engine


def fetch_samples():
    """Download same samples to evaluate and add to session."""
    st.session_state.dag_name = st.session_state.get("dag_name")
    if not st.session_state.dag_name_input:
        raise RuntimeError("A DAG name must be set in environment variables.")

    dag = DAG(st.session_state.dag_name)

    with st.spinner("Loading samples..."):
        st.session_state.samples = get_samples(
            dag=dag,
            n=100,
            user_id=st.session_state.user_id,
            default_client=init_engine(),
        )
    st.session_state.step = "eval" if st.session_state.samples else "done"


st.title("Matchbox evaluation session")

if "step" not in st.session_state:
    if not settings.user:
        raise MatchboxClientSettingsException("User name is unset.")
    st.session_state.user_name = settings.user
    st.session_state.user_id = _handler.login(user_name=st.session_state.user_name)
    st.session_state.step = "ready"


if st.session_state.step == "ready":
    st.text_input("DAG to sample from", key="dag_name_input")
    st.button("Fetch samples", type="primary", on_click=fetch_samples)


if st.session_state.step == "eval":
    if "df" not in st.session_state:
        st.session_state.shown_cluster = next(iter(st.session_state.samples.keys()))

        original_df = st.session_state.samples.pop(st.session_state.shown_cluster)

        st.session_state.judgement = []
        st.session_state.orig_columns = list(set(original_df.columns) - {"leaf"})
        st.session_state.df = original_df.with_columns(select=False)

    st.markdown(
        f"Welcome **{st.session_state.user_name}**. "
        f"Sampling from DAG: `{st.session_state.dag_name}`"
    )
    edited_df = st.data_editor(
        st.session_state.df,
        disabled=st.session_state.orig_columns,
        column_order=("select", *st.session_state.orig_columns),
    )

    def splinter():
        """Endorse subset of cluster."""
        st.session_state.df = edited_df.filter(~pl.col("select"))
        selected_ids = (
            edited_df.filter(pl.col("select")).select("leaf").to_series().to_list()
        )
        st.session_state.judgement.append(selected_ids)
        if st.session_state.df.shape[0] == 1:
            looks_good()

    def looks_good():
        """Confirm cluster as is."""
        all_ids = st.session_state.df.select("leaf").to_series().to_list()
        st.session_state.judgement.append(all_ids)
        _handler.send_eval_judgement(
            judgement=Judgement(
                shown=st.session_state.shown_cluster,
                endorsed=st.session_state.judgement,
                user_id=st.session_state.user_id,
            ),
        )

        del st.session_state.df

        if not len(st.session_state.samples):
            st.session_state.step = "done"

    if (edited_df.select("select").to_series().any()) and (
        not edited_df.select("select").to_series().all()
    ):
        st.button("Splinter", icon="✂️", on_click=splinter)
    else:
        st.button("Looks good to me.", icon="✅", on_click=looks_good)

if st.session_state.step == "done":
    st.header("You're all done 🔥")



"""A linking methodology based on a deterministic set of conditions."""

import polars as pl
from pydantic import Field, field_validator

from matchbox.client.models import comparison
from matchbox.client.models.linkers.base import Linker, LinkerSettings


class DeterministicSettings(LinkerSettings):
    """A data class to enforce the Deterministic linker's settings dictionary shape."""

    comparisons: list[str] = Field(
        description="""
            A list of valid ON clause to compare fields between the left and 
            the right data.

            Use left.field and right.field to refer to columns in the respective 
            sources.

            Each comparison will be treated as OR logic, but more efficiently than using
            an OR condition in the SQL WHERE clause.

            For example:

            [   
                "left.company_number = right.company_number",
                "left.name = right.name and left.postcode = right.postcode",
            ]
        """,
    )

    @field_validator("comparisons", mode="before")
    @classmethod
    def validate_comparison(cls, value: str | list[str]) -> list[str]:
        """Turn single string into list of one string."""
        if isinstance(value, str):
            return [comparison(value)]
        return [comparison(v) for v in value]


class DeterministicLinker(Linker):
    """A deterministic linker that links based on a set of boolean conditions."""

    settings: DeterministicSettings

    def prepare(self, left: pl.DataFrame, right: pl.DataFrame) -> None:
        """Prepare the linker for linking."""
        pass

    def link(self, left: pl.DataFrame, right: pl.DataFrame) -> pl.DataFrame:
        """Link the left and right dataframes."""
        left_pl = left.lazy()  # noqa: F841
        right_pl = right.lazy()  # noqa: F841

        subqueries = []
        for i, condition in enumerate(self.settings.comparisons):
            subquery = f"""
                SELECT
                    l_{i}.{self.settings.left_id} AS left_id,
                    r_{i}.{self.settings.right_id} AS right_id,
                    1.0 AS probability
                FROM left_pl l_{i}
                INNER JOIN right_pl r_{i}
                    ON {condition}
            """
            subqueries.append(subquery)

        union_query = " UNION ALL ".join(subqueries)

        final_query = f"""
            SELECT DISTINCT 
                final.left_id, 
                final.right_id, 
                final.probability
            FROM ({union_query}) as final
        """

        return (
            pl.sql(final_query)
            .with_columns(
                [
                    pl.col("probability").cast(pl.Float32),
                ]
            )
            .collect()
        )



"""Utilities for JWT API authorisation.

NOTE: A secure set-up requires the JWT generation logic to not live on the client.
Instead, client JWT should be generated within a secure environment with access to
the private key.

This file supports the edge case of a trusted client bypassing the JWT mechanism
altogether, which is not recommended in general.

"""

import json
import time
from base64 import urlsafe_b64encode

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PrivateKey
from cryptography.hazmat.primitives.serialization import load_pem_private_key

from matchbox.client._settings import settings

EXPIRY_AFTER_X_HOURS = 24


def b64encode_nopadding(to_encode):
    """B64 encode."""
    return urlsafe_b64encode(to_encode).rstrip(b"=")


def generate_json_web_token(sub):
    """Generate JWT with private API Key."""
    private_key = load_pem_private_key(
        settings.private_key.get_secret_value().encode(), password=None
    )
    header = {
        "typ": "JWT",
        "alg": "EdDSA",
        "crv": "Ed25519",
    }
    payload = {
        "sub": sub,
        "exp": int(time.time() + 60 * 60 * EXPIRY_AFTER_X_HOURS),
        "authorised_hosts": settings.api_root,
    }
    to_sign = (
        b64encode_nopadding(json.dumps(header).encode("utf-8"))
        + b"."
        + b64encode_nopadding(json.dumps(payload).encode("utf-8"))
    )
    signature = b64encode_nopadding(private_key.sign(to_sign))
    token = (to_sign + b"." + signature).decode()
    return token


def generate_EdDSA_key_pair() -> tuple[bytes, bytes]:
    """Generate private and public key pair."""
    private_key = Ed25519PrivateKey.generate()

    unencrypted_pem_private_key = private_key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )

    pem_public_key = private_key.public_key().public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    return unencrypted_pem_private_key, pem_public_key



"""Definition of model inputs."""

import warnings
from datetime import datetime
from typing import TYPE_CHECKING, Any, Self

import duckdb
import polars as pl
from polars import DataFrame as PolarsDataFrame
from sqlglot import expressions, parse_one
from sqlglot import select as sqlglot_select

from matchbox.client import _handler
from matchbox.client.models import Model
from matchbox.client.models.dedupers.base import Deduper, DeduperSettings
from matchbox.client.models.linkers.base import Linker, LinkerSettings
from matchbox.client.sources import Source
from matchbox.common.db import QueryReturnClass, QueryReturnType
from matchbox.common.dtos import (
    QueryCombineType,
    QueryConfig,
)

if TYPE_CHECKING:
    from matchbox.client.dags import DAG
else:
    DAG = Any


class Query:
    """Queriable input to a model."""

    def __init__(
        self,
        *sources: Source,
        dag: DAG,
        model: Model | None = None,
        combine_type: QueryCombineType = QueryCombineType.CONCAT,
        threshold: float | None = None,
        cleaning: dict[str, str] | None = None,
    ):
        """Initialise query.

        Args:
            sources: List of sources to query from
            dag: DAG containing sources and models.
            model (optional): Model to use to resolve sources. It can only be missing
                if querying from a single source.
            combine_type (optional): How to combine the data from different sources.
                Default is `concat`.

                * If `concat`, concatenate all sources queried without any merging.
                    Multiple rows per ID, with null values where data isn't available
                * If `explode`, outer join on Matchbox ID. Multiple rows per ID,
                    with one for every unique combination of data requested
                    across all sources
                * If `set_agg`, join on Matchbox ID, group on Matchbox ID, then
                    aggregate to nested lists of unique values. One row per ID,
                    but all requested data is in nested arrays

            threshold (optional): The threshold to use for creating clusters
                If None, uses the resolutions' default threshold
                If an integer, uses that threshold for the specified resolution, and the
                resolution's cached thresholds for its ancestors

            cleaning (optional): A dictionary mapping an output column name to a SQL
                expression that will populate a new column.
        """
        self.last_run: datetime | None = None
        self.raw_data: PolarsDataFrame | None = None
        self.dag = dag
        self.sources = sources
        self.model = model
        self.config = QueryConfig(
            source_resolutions=[source.name for source in sources],
            model_resolution=model.name if model else None,
            combine_type=combine_type,
            threshold=int(threshold * 100) if threshold else None,
            cleaning=cleaning,
        )

    def run(
        self,
        return_type: QueryReturnType = QueryReturnType.POLARS,
        return_leaf_id: bool = False,
        batch_size: int | None = None,
        full_rerun: bool = False,
        cache_raw: bool = False,
    ) -> QueryReturnClass:
        """Runs queries against the selected backend.

        Args:
            return_type (optional): Type of dataframe returned, defaults to "polars".
                Other options are "pandas" and "arrow".
            return_leaf_id (optional): Whether matchbox IDs for source clusters should
                be saved as a byproduct in the `leaf_ids` attribute.
            batch_size (optional): The size of each batch when fetching data from the
                warehouse, which helps reduce memory usage and load on the database.
                Default is None.
            full_rerun: Whether to force a re-run of the query
            cache_raw: Whether to store the pre-cleaned data to iterate on cleaning.

        Returns: Data in the requested return type

        Raises:
            MatchboxEmptyServerResponse: If no data was returned by the server.
        """
        if self.last_run and not full_rerun:
            warnings.warn("Query already run, skipping.", UserWarning, stacklevel=2)
            return self.data

        source_results: list[PolarsDataFrame] = []
        for source in self.sources:
            mb_ids = pl.from_arrow(
                _handler.query(
                    source=source.name,
                    resolution=self.model.name if self.model else None,
                    threshold=self.config.threshold,
                    return_leaf_id=return_leaf_id,
                )
            )

            raw_batches = source.fetch(
                qualify_names=True,
                batch_size=batch_size,
                return_type=QueryReturnType.POLARS,
            )

            processed_batches = [
                b.join(
                    other=mb_ids,
                    left_on=source.qualified_key,
                    right_on="key",
                    how="inner",
                )
                for b in raw_batches
            ]
            source_results.append(pl.concat(processed_batches, how="vertical"))

        # Process all data and return a single result
        tables: list[PolarsDataFrame] = list(source_results)

        # Combine results based on combine_type
        if return_leaf_id:
            concatenated = pl.concat(tables, how="diagonal")
            self.leaf_id = concatenated.select(["id", "leaf_id"])

        if self.config.combine_type == QueryCombineType.CONCAT:
            if return_leaf_id:  # can reuse the concatenated dataframe
                raw_data = concatenated.drop(["leaf_id"])
            else:
                raw_data = pl.concat(tables, how="diagonal")
        else:
            raw_data = tables[0].drop("leaf_id", strict=False)
            for table in tables[1:]:
                raw_data = raw_data.join(
                    table.drop("leaf_id", strict=False),
                    on="id",
                    how="full",
                    coalesce=True,
                )

            raw_data = raw_data.select(["id", pl.all().exclude("id")])

            if self.config.combine_type == QueryCombineType.SET_AGG:
                # Aggregate into lists
                agg_expressions = [
                    pl.col(col).unique() for col in raw_data.columns if col != "id"
                ]
                raw_data = raw_data.group_by("id").agg(agg_expressions)

        if cache_raw:
            self.raw_data = raw_data
        clean_data = _convert_df(
            clean(raw_data, self.config.cleaning), return_type=return_type
        )

        self.data = clean_data
        self.last_run = datetime.now()

        return self.data

    def clean(
        self,
        cleaning: dict[str, str] | None,
        return_type: QueryReturnType = QueryReturnType.POLARS,
    ) -> QueryReturnClass:
        """Change cleaning dictionary and re-apply cleaning, if raw data was cached.

        Args:
            cleaning: A dictionary mapping field aliases to SQL expressions.
                The SQL expressions can reference columns in the data using their names.
                If None, no cleaning is applied and the original data is returned.
                `SourceConfig.f()` can be used to help reference qualified fields.
            return_type (optional): Type of dataframe returned, defaults to "polars".
                    Other options are "pandas" and "arrow".
        """
        if self.raw_data is None:
            raise RuntimeError("No raw data is stored in this query.")

        self.config = self.config.model_copy(update={"cleaning": cleaning})

        self.data = _convert_df(
            data=clean(data=self.raw_data, cleaning_dict=cleaning),
            return_type=return_type,
        )
        return self.data

    def deduper(
        self,
        name: str,
        model_class: Deduper,
        model_settings: DeduperSettings,
        description: str | None = None,
    ) -> Model:
        """Create deduper for data in this query."""
        return self.dag.model(
            name=name,
            description=description,
            model_class=model_class,
            model_settings=model_settings,
            left_query=self,
        )

    def linker(
        self,
        other_query: Self,
        name: str,
        model_class: Linker,
        model_settings: LinkerSettings,
        description: str | None = None,
    ) -> Model:
        """Create linker for data in this query and another query."""
        return self.dag.model(
            name=name,
            description=description,
            model_class=model_class,
            model_settings=model_settings,
            left_query=self,
            right_query=other_query,
        )


def clean(data: pl.DataFrame, cleaning_dict: dict[str, str] | None) -> pl.DataFrame:
    """Clean data using DuckDB with the provided cleaning SQL.

    * ID is passed through automatically
    * If present, leaf_id and key are passed through automatically
    * Columns not mentioned in the cleaning_dict are passed through unchanged
    * Each key in cleaning_dict is an alias for a SQL expression

    Args:
        data: Raw polars dataframe to clean
        cleaning_dict: A dictionary mapping field aliases to SQL expressions.
            The SQL expressions can reference columns in the data using their names.
            If None, no cleaning is applied and the original data is returned.
            `SourceConfig.f()` can be used to help reference qualified fields.

    Returns:
        Cleaned polars dataframe

    Examples:
        Column passthrough behavior:

        ```python
        data = pl.DataFrame(
            {
                "id": [1, 2, 3],
                "name": ["John", "Jane", "Bob"],
                "age": [25, 30, 35],
                "city": ["London", "Hull", "Stratford-upon-Avon"],
            }
        )
        cleaning_dict = {
            "full_name": "name"  # Only references 'name' column
        }
        result = clean(data, cleaning_dict)
        # Result columns: id, full_name, age, city
        # 'name' is dropped because it was used in cleaning_dict
        # 'age' and 'city' are passed through unchanged
        ```

        Multiple column references:

        ```python
        data = pl.DataFrame(
            {
                "id": [1, 2, 3],
                "first": ["John", "Jane", "Bob"],
                "last": ["Doe", "Smith", "Johnson"],
                "salary": [50000, 60000, 55000],
            }
        )
        cleaning_dict = {
            "name": "first || ' ' || last",  # References both 'first' and 'last'
            "high_earner": "salary > 55000",
        }
        result = clean(data, cleaning_dict)
        # Result columns: id, name, high_earner
        # 'first', 'last', and 'salary' are dropped (used in expressions)
        # No other columns to pass through
        ```

        Special columns (leaf_id, key) handling:

        ```python
        data = pl.DataFrame(
            {
                "id": [1, 2, 3],
                "leaf_id": ["a", "b", "c"],
                "key": ["x", "y", "z"],
                "value": [10, 20, 30],
                "status": ["active", "inactive", "pending"],
            }
        )
        cleaning_dict = {"processed_value": "value * 2"}
        result = clean(data, cleaning_dict)
        # Result columns: id, leaf_id, key, processed_value, status
        # 'id', 'leaf_id', 'key' always included automatically
        # 'value' dropped (used in expression), 'status' passed through
        ```

        No cleaning (returns original data):

        ```python
        data = pl.DataFrame({"id": [1, 2], "name": ["Alice", "Bob"], "score": [95, 87]})
        result = clean(data, None)
        # Returns exact same dataframe with all original columns
        ```
    """
    if cleaning_dict is None:
        return data

    def _add_column(name: str) -> expressions.Alias:
        """Select and alias a column with the given name."""
        return expressions.Alias(
            this=expressions.Column(
                this=expressions.Identifier(this=name, quoted=False)
            ),
            alias=expressions.Identifier(this=name, quoted=False),
        )

    # Always select 'id'
    to_select: list[expressions.Expression] = [_add_column("id")]

    # Add optional columns if they exist
    for col in ["leaf_id", "key"]:
        if col in data.columns:
            to_select.append(_add_column(col))

    # Parse and add each SQL expression from cleaning_dict
    query_column_names: set[str] = set()
    for alias, sql in cleaning_dict.items():
        stmt = parse_one(sql, dialect="duckdb")

        # Get column name used in the expression
        for node in stmt.walk():
            if isinstance(node, expressions.Column):
                query_column_names.add(node.name)

        # Add to the list of expressions to select
        to_select.append(expressions.alias_(stmt, alias))

    # Add all column names not used in the query
    for column in set(data.columns) - query_column_names - {"id", "leaf_id", "key"}:
        to_select.append(
            expressions.Alias(
                this=expressions.Column(
                    this=expressions.Identifier(this=column, quoted=False)
                ),
                alias=expressions.Identifier(this=column, quoted=False),
            )
        )

    query = sqlglot_select(*to_select, dialect="duckdb").from_("data")

    with duckdb.connect(":memory:") as conn:
        conn.register("data", data)
        return conn.execute(query.sql(dialect="duckdb")).pl()


def _convert_df(data: PolarsDataFrame, return_type: QueryReturnType):
    match return_type:
        case QueryReturnType.POLARS:
            return data
        case QueryReturnType.PANDAS:
            return data.to_pandas()
        case QueryReturnType.ARROW:
            return data.to_arrow()
        case _:
            raise ValueError(f"Return type {return_type} is invalid")



from collections.abc import Callable

import polars as pl
import pytest
from httpx import Response
from polars.testing import assert_frame_equal
from pyarrow import Table
from respx import MockRouter
from sqlalchemy import Engine

from matchbox.client.dags import DAG
from matchbox.client.eval import get_samples
from matchbox.client.models.linkers import DeterministicLinker
from matchbox.common.arrow import SCHEMA_EVAL_SAMPLES, table_to_buffer
from matchbox.common.exceptions import MatchboxSourceTableError
from matchbox.common.factories.sources import source_from_tuple


def test_get_samples(
    matchbox_api: MockRouter,
    sqlite_warehouse: Engine,
    sqlite_in_memory_warehouse: Engine,
    env_setter: Callable[[str, str], None],
):
    # Make dummmy data
    user_id = 12

    # Foo has two identical rows
    foo_testkit = source_from_tuple(
        data_tuple=({"col": 1}, {"col": 1}, {"col": 2}, {"col": 3}, {"col": 4}),
        data_keys=["1", "1bis", "2", "3", "4"],
        name="foo",
        location_name="db",
        engine=sqlite_warehouse,
    ).write_to_location()

    bar_testkit = source_from_tuple(
        data_tuple=({"col": 1}, {"col": 2}, {"col": 3}, {"col": 4}),
        data_keys=["a", "b", "c", "d"],
        name="bar",
        location_name="db",
        engine=sqlite_warehouse,
    ).write_to_location()

    # This will be excluded as the location name differs
    baz_testkit = source_from_tuple(
        data_tuple=({"col": 1},),
        data_keys=["x"],
        name="baz",
        location_name="db_other",
        engine=sqlite_warehouse,
    ).write_to_location()

    dag = DAG("companies", new=True)
    foo = dag.source(**foo_testkit.into_dag())
    bar = dag.source(**bar_testkit.into_dag())
    baz = dag.source(**baz_testkit.into_dag())
    foo.query().linker(
        bar.query(),
        name="linker1",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.key=r.key"},
    ).query(foo, bar).linker(
        baz.query(),
        name="linker2",
        model_class=DeterministicLinker,
        model_settings={"comparisons": "l.key=r.key"},
    )

    # Mock API endpoints
    matchbox_api.get("/resolutions/foo").mock(
        return_value=Response(
            200, content=foo_testkit.source.to_resolution().model_dump_json()
        )
    )
    matchbox_api.get("/resolutions/bar").mock(
        return_value=Response(
            200, content=bar_testkit.source.to_resolution().model_dump_json()
        )
    )
    matchbox_api.get("/resolutions/baz").mock(
        return_value=Response(
            200, content=baz_testkit.source.to_resolution().model_dump_json()
        )
    )

    # Mock samples
    samples = Table.from_pylist(
        [
            # Source foo - with two keys for one leaf
            {"root": 10, "leaf": 1, "key": "1", "source": "foo"},
            {"root": 10, "leaf": 1, "key": "1bis", "source": "foo"},
            {"root": 10, "leaf": 2, "key": "2", "source": "foo"},
            {"root": 11, "leaf": 3, "key": "3", "source": "foo"},
            {"root": 11, "leaf": 4, "key": "4", "source": "foo"},
            # Source bar
            {"root": 10, "leaf": 5, "key": "a", "source": "bar"},
            {"root": 10, "leaf": 6, "key": "b", "source": "bar"},
            {"root": 11, "leaf": 7, "key": "c", "source": "bar"},
            {"root": 11, "leaf": 8, "key": "d", "source": "bar"},
            # Source baz
            {"root": 10, "leaf": 1, "key": "x", "source": "baz"},
        ],
        schema=SCHEMA_EVAL_SAMPLES,
    )
    # There will be nulls in case of a schema mismatch
    assert len(samples.drop_null()) == len(samples)

    matchbox_api.get("/eval/samples").mock(
        return_value=Response(200, content=table_to_buffer(samples).read())
    )

    # Check results
    with pytest.warns(UserWarning, match="Skipping"):
        samples = get_samples(
            n=10,
            dag=dag,
            user_id=user_id,
            clients={"db": sqlite_warehouse},
        )

    assert sorted(samples.keys()) == [10, 11]

    expected_sample_10 = pl.DataFrame(
        {
            "leaf": [1, 1, 2, 5, 6],
            "key": ["1", "1bis", "2", "a", "b"],
            "foo_col": [1, 1, 2, None, None],
            "bar_col": [None, None, None, 1, 2],
        }
    )

    expected_sample_11 = pl.DataFrame(
        {
            "leaf": [3, 4, 7, 8],
            "key": ["3", "4", "c", "d"],
            "foo_col": [3, 4, None, None],
            "bar_col": [None, None, 3, 4],
        }
    )

    assert_frame_equal(
        samples[10],
        expected_sample_10,
        check_column_order=False,
        check_row_order=False,
        check_dtypes=False,
    )
    assert_frame_equal(
        samples[11],
        expected_sample_11,
        check_column_order=False,
        check_row_order=False,
        check_dtypes=False,
    )

    # What happens if no samples are available?
    matchbox_api.get("/eval/samples").mock(
        return_value=Response(
            200,
            content=table_to_buffer(
                Table.from_pylist([], schema=SCHEMA_EVAL_SAMPLES)
            ).read(),
        )
    )

    no_samples = get_samples(
        n=10,
        dag=dag,
        user_id=user_id,
        clients={"db": sqlite_warehouse},
    )
    assert no_samples == {}

    # And if no client available?
    just_baz_samples = Table.from_pylist(
        [{"root": 10, "leaf": 1, "key": "x", "source": "baz"}],
        schema=SCHEMA_EVAL_SAMPLES,
    )
    matchbox_api.get("/eval/samples").mock(
        return_value=Response(
            200,
            content=table_to_buffer(just_baz_samples).read(),
        )
    )
    with pytest.warns(
        UserWarning, match="Skipping baz, incompatible with given client"
    ):
        no_accessible_samples = get_samples(n=10, dag=dag, user_id=user_id)
    assert no_accessible_samples == {}

    # Using default client as fallback
    samples_default_creds = get_samples(
        n=10,
        dag=dag,
        user_id=user_id,
        default_client=sqlite_warehouse,
    )
    assert len(samples_default_creds) == 1

    # What happens if source cannot be queried using client?
    with pytest.raises(MatchboxSourceTableError):
        get_samples(
            n=10,
            dag=dag,
            user_id=user_id,
            default_client=sqlite_in_memory_warehouse,
        )



"""Test deterministic behavior of dedupers."""

from collections.abc import Callable
from typing import Any
from unittest.mock import Mock, patch

import polars as pl
import pytest

from matchbox.client.models import Model
from matchbox.client.models.dedupers.base import Deduper
from matchbox.client.models.dedupers.naive import NaiveDeduper, NaiveSettings
from matchbox.client.queries import Query
from matchbox.client.results import Results
from matchbox.common.factories.entities import FeatureConfig
from matchbox.common.factories.sources import (
    SourceTestkit,
    SourceTestkitParameters,
    linked_sources_factory,
)

DeduperConfigurator = Callable[[SourceTestkit], dict[str, Any]]

# Methodology configuration adapters


def configure_naive_deduper(testkit: SourceTestkit) -> dict[str, Any]:
    """Configure settings for NaiveDeduper.

    Args:
        testkit: SourceTestkit object from linked_sources_factory

    Returns:
        A dictionary with validated settings for NaiveDeduper
    """
    # Extract field names excluding key and id
    fields = [
        c.name
        for c in testkit.source_config.index_fields
        if c.name not in ("key", "id")
    ]

    settings_dict = {
        "id": "id",
        "unique_fields": fields,
    }

    NaiveSettings.model_validate(settings_dict)

    return settings_dict


DEDUPERS = [
    pytest.param(NaiveDeduper, configure_naive_deduper, id="Naive"),
    # Add more deduper classes and configuration functions here
]


# Test cases


@pytest.mark.parametrize(("Deduper", "configure_deduper"), DEDUPERS)
@patch.object(Query, "run")
def test_no_deduplication(
    mock_query_run: Mock,
    Deduper: Deduper,
    configure_deduper: DeduperConfigurator,
):
    """Test deduplication where there aren't actually any duplicates."""
    # Create a source with exact duplicates
    features = (
        FeatureConfig(
            name="company",
            base_generator="company",
        ),
        FeatureConfig(
            name="email",
            base_generator="email",
        ),
    )

    source_parameters = SourceTestkitParameters(
        name="source_exact",
        features=features,
        n_true_entities=10,
        repetition=0,  # Each entity appears once
    )

    linked = linked_sources_factory(source_parameters=(source_parameters,), seed=42)
    source_testkit = linked.sources["source_exact"]

    mock_query_run.return_value = pl.from_arrow(source_testkit.data)

    # Mock query to server

    # Configure and run the deduper
    deduper = Model(
        dag=linked.dag,
        name="exact_deduper",
        description="Deduplication of exact duplicates",
        model_class=Deduper,
        model_settings=configure_deduper(source_testkit),
        left_query=Query(source_testkit.source, dag=linked.dag),
    )
    results: Results = deduper.run()

    # Validate results against ground truth
    identical, report = linked.diff_results(
        probabilities=results.probabilities,
        left_clusters=source_testkit.entities,
        right_clusters=None,
        sources=["source_exact"],
        threshold=0,
    )

    assert identical, f"Expected perfect results but got: {report}"


@pytest.mark.parametrize(("Deduper", "configure_deduper"), DEDUPERS)
@patch.object(Query, "run")
def test_exact_duplicate_deduplication(
    mock_query_run: Mock, Deduper: Deduper, configure_deduper: DeduperConfigurator
):
    """Test deduplication with exact duplicates."""
    # Create a source with exact duplicates
    features = (
        FeatureConfig(
            name="company",
            base_generator="company",
        ),
        FeatureConfig(
            name="email",
            base_generator="email",
        ),
    )

    source_parameters = SourceTestkitParameters(
        name="source_exact",
        features=features,
        n_true_entities=10,
        repetition=2,  # Each entity appears 3 times (base + 2 repetitions)
    )

    linked = linked_sources_factory(source_parameters=(source_parameters,), seed=42)
    source = linked.sources["source_exact"]

    mock_query_run.return_value = pl.from_arrow(source.data)

    # Configure and run the deduper
    deduper = Model(
        dag=linked.dag,
        name="exact_deduper",
        description="Deduplication of exact duplicates",
        model_class=Deduper,
        model_settings=configure_deduper(source),
        left_query=Query(source.source, dag=linked.dag),
    )
    results: Results = deduper.run()

    # Validate results against ground truth
    identical, report = linked.diff_results(
        probabilities=results.probabilities,
        left_clusters=source.entities,
        right_clusters=None,
        sources=["source_exact"],
        threshold=0,
    )

    assert identical, f"Expected perfect results but got: {report}"



"""General utilities for the PostgreSQL backend."""

import base64
import contextlib
import cProfile
import io
import pstats
import uuid
from collections.abc import Generator

import pyarrow as pa
from adbc_driver_manager import ProgrammingError as ADBCProgrammingError
from adbc_driver_postgresql.dbapi import Connection as ADBCConnection
from pyarrow import Table as ArrowTable
from sqlalchemy import Column, MetaData, Table, func, select, text
from sqlalchemy.dialects import postgresql
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.orm import DeclarativeMeta
from sqlalchemy.pool import PoolProxiedConnection
from sqlalchemy.sql import Select
from sqlalchemy.sql.type_api import TypeEngine

from matchbox.common.exceptions import (
    MatchboxDatabaseWriteError,
)
from matchbox.common.graph import (
    ResolutionEdge,
    ResolutionGraph,
    ResolutionNode,
    ResolutionType,
)
from matchbox.common.logging import logger
from matchbox.server.base import MatchboxBackends, MatchboxSnapshot
from matchbox.server.postgresql.db import MBDB
from matchbox.server.postgresql.orm import ResolutionFrom, Resolutions

# Retrieval


def get_resolution_graph() -> ResolutionGraph:
    """Retrieves the resolution graph."""
    G = ResolutionGraph(nodes=set(), edges=set())
    with MBDB.get_session() as session:
        for resolution in session.query(Resolutions).all():
            G.nodes.add(
                ResolutionNode(
                    id=resolution.resolution_id,
                    name=resolution.name,
                    type=ResolutionType(resolution.type),
                )
            )

        for edge in (
            session.query(ResolutionFrom).filter(ResolutionFrom.level == 1).all()
        ):
            G.edges.add(ResolutionEdge(parent=edge.parent, child=edge.child))

    return G


# Data management


def dump() -> MatchboxSnapshot:
    """Dumps the entire database to a snapshot.

    Returns:
        A MatchboxSnapshot object of type "postgres" with the database's
            current state.
    """
    data = {}

    with MBDB.get_session() as session:
        for table in MBDB.sorted_tables:
            # Query all records from this table
            records = session.execute(select(table)).mappings().all()

            # Convert each record to a dictionary
            table_data = []
            for record in records:
                record_dict = dict(record)
                for k, v in record_dict.items():
                    # Store bytes as nested dictionary with encoding format
                    if isinstance(v, bytes):
                        record_dict[k] = {"base64": base64.b64encode(v).decode("ascii")}

                table_data.append(record_dict)

            data[table.name] = table_data

    return MatchboxSnapshot(backend_type=MatchboxBackends.POSTGRES, data=data)


def restore(snapshot: MatchboxSnapshot, batch_size: int) -> None:
    """Restores the database from a snapshot.

    Args:
        snapshot: A MatchboxSnapshot object of type "postgres" with the
            database's state
        batch_size: The number of records to insert in each batch

    Raises:
        ValueError: If the snapshot is missing data
    """
    with MBDB.get_session() as session:
        # Process tables in order
        for table in MBDB.sorted_tables:
            if table.name not in snapshot.data:
                raise ValueError(f"Invalid: Table {table.name} not found in snapshot.")

            records = snapshot.data[table.name]

            if not records:
                continue

            # Process records for insertion
            processed_records = []
            for record in records:
                processed_record = {}

                for key, value in record.items():
                    # Check if the value is a dictionary with encoding format
                    if isinstance(value, dict) and "base64" in value:
                        processed_record[key] = base64.b64decode(value["base64"])
                    else:
                        processed_record[key] = value

                processed_records.append(processed_record)

            # Insert
            for i in range(0, len(processed_records), batch_size):
                batch = processed_records[i : i + batch_size]
                session.execute(insert(table), batch)
                session.flush()

            # Re-sync primary key sequences
            for pk in table.primary_key:
                lock_stmt = text(
                    f"lock table {table.schema}.{table.name} in exclusive mode;"
                )
                sync_statement = select(
                    func.setval(
                        func.pg_get_serial_sequence(
                            text(f"'{table.schema}.{table.name}'"),
                            text(f"'{pk.name}'"),
                        ),
                        func.coalesce(func.max(text(pk.name)), 0),
                    )
                ).select_from(text(f"{table.schema}.{table.name}"))

                session.execute(lock_stmt)
                session.execute(sync_statement)

        session.commit()


# SQLAlchemy profiling


@contextlib.contextmanager
def sqa_profiled():
    """SQLAlchemy profiler.

    Taken directly from their docs:
    https://docs.sqlalchemy.org/en/20/faq/performance.html#query-profiling
    """
    pr = cProfile.Profile()
    pr.enable()
    yield
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats("cumulative")
    ps.print_stats()
    # uncomment this to see who's calling what
    # ps.print_callers()
    print(s.getvalue())


# Misc


def compile_sql(stmt: Select) -> str:
    """Compiles a SQLAlchemy statement into a string.

    Args:
        stmt: The SQLAlchemy statement to compile.

    Returns:
        The compiled SQL statement as a string.
    """
    return str(
        stmt.compile(
            dialect=postgresql.dialect(), compile_kwargs={"literal_binds": True}
        )
    )


def _copy_to_table(
    table_name: str,
    schema_name: str,
    connection: ADBCConnection,
    data: ArrowTable,
    max_chunksize: int | None = None,
):
    """Copy data to table using ADBC with isolated connection."""
    batch_reader = pa.RecordBatchReader.from_batches(
        data.schema, data.to_batches(max_chunksize=max_chunksize)
    )

    with connection.cursor() as cursor:
        cursor.adbc_ingest(
            table_name=table_name,
            data=batch_reader,
            mode="append",
            db_schema_name=schema_name,
        )


def large_append(
    data: pa.Table,
    table_class: DeclarativeMeta,
    adbc_connection: PoolProxiedConnection,
    max_chunksize: int | None = None,
):
    """Append a PyArrow table to a PostgreSQL table using ADBC.

    This function does not support upserting and will error if keys clash.
    This method does not auto-commit, which is the responsibility of the caller.

    Args:
        data: A PyArrow table to write.
        table_class: The SQLAlchemy ORM class for the table to write to.
        adbc_connection: An ADBC connection from the pool.
            This is returned by MBDB.get_adbc_connection() and needs to be used via a
            context manager.
        max_chunksize: Size of data chunks to be read and copied.
    """
    table: Table = table_class.__table__

    table_columns = [c.name for c in table.columns]
    col_diff = set(data.column_names) - set(table_columns)
    if len(col_diff) > 0:
        raise ValueError(f"Table {table.name} does not have columns {col_diff}")
    try:
        _copy_to_table(
            table_name=table.name,
            schema_name=table.schema,
            connection=adbc_connection,
            data=data,
            max_chunksize=max_chunksize,
        )
    except ADBCProgrammingError as e:
        raise MatchboxDatabaseWriteError from e


@contextlib.contextmanager
def ingest_to_temporary_table(
    table_name: str,
    schema_name: str,
    data: ArrowTable,
    column_types: dict[str, type[TypeEngine]],
    max_chunksize: int | None = None,
) -> Generator[Table, None, None]:
    """Context manager to ingest Arrow data to a temporary table with explicit types.

    Args:
        table_name: Base name for the temporary table
        schema_name: Schema where the temporary table will be created
        data: PyArrow table containing the data to ingest
        column_types: Map of column names to SQLAlchemy types
        max_chunksize: Optional maximum chunk size for batches

    Returns:
        A SQLAlchemy Table object representing the temporary table
    """
    temp_table_name = f"{table_name}_tmp_{uuid.uuid4().hex}"

    # Validate that all data columns have type mappings
    missing_columns = set(data.column_names) - set(column_types.keys())
    if missing_columns:
        raise ValueError(f"Missing type mappings for columns: {missing_columns}")

    try:
        # Create SQLAlchemy Table from explicit type mapping
        metadata = MetaData(schema=schema_name)
        columns = [
            Column(column_name, column_type())
            for column_name, column_type in column_types.items()
            if column_name in data.column_names
        ]
        temp_table = Table(temp_table_name, metadata, *columns)

        with MBDB.get_session() as session:
            temp_table.create(session.bind)
            session.commit()

        # Ingest data into the temporary table
        with MBDB.get_adbc_connection() as connection:
            _copy_to_table(
                table_name=temp_table_name,
                schema_name=schema_name,
                connection=connection,
                data=data,
                max_chunksize=max_chunksize,
            )
            connection.commit()

        yield temp_table

    finally:
        # Step 3: Clean up
        try:
            with MBDB.get_session() as session:
                temp_table.drop(session.bind, checkfirst=True)
                session.commit()
        except Exception as e:  # noqa: BLE001
            logger.warning(f"Failed to drop temp table {temp_table_name}: {e}")



"""Common data structures for resolution graphs."""

from enum import StrEnum
from typing import TypeAlias

import rustworkx as rx
from pydantic import BaseModel, ConfigDict

SourceResolutionName: TypeAlias = str
"""Type alias for source resolution names."""

ModelResolutionName: TypeAlias = str
"""Type alias for model resolution names."""

ResolutionName = SourceResolutionName | ModelResolutionName
"""Type alias for resolution names."""

DEFAULT_RESOLUTION: ResolutionName = "__DEFAULT__"


class ResolutionType(StrEnum):
    """Types of nodes in a resolution."""

    SOURCE = "source"
    MODEL = "model"


class ResolutionNode(BaseModel):
    """A node in a resolution graph."""

    model_config = ConfigDict(frozen=True)

    id: int
    name: ResolutionName
    type: ResolutionType


class ResolutionEdge(BaseModel):
    """An edge in a resolution graph."""

    model_config = ConfigDict(frozen=True)

    parent: int
    child: int


class ResolutionGraph(BaseModel):
    """A directed graph of resolution nodes and edges."""

    nodes: set[ResolutionNode]
    edges: set[ResolutionEdge]

    def to_rx(self) -> rx.PyDiGraph:
        """Convert the resolution graph to a rustworkx directed graph."""
        nodes = {}
        G = rx.PyDiGraph()
        for n in self.nodes:
            node_data = {
                "id": n.id,
                "name": n.name,
                "type": str(n.type),
            }
            nodes[n.id] = G.add_node(node_data)
        for e in self.edges:
            G.add_edge(nodes[e.parent], nodes[e.child], {})
        return G



"""Interface to locations where source data is stored."""

import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable, Generator, Iterable, Iterator
from contextlib import contextmanager
from copy import deepcopy
from datetime import datetime
from functools import wraps
from typing import TYPE_CHECKING, Any, ParamSpec, Self, TypeVar, overload

import polars as pl
import sqlglot
from pyarrow import Table as ArrowTable
from sqlalchemy import Engine
from sqlalchemy.exc import OperationalError

from matchbox.client import _handler
from matchbox.common.db import (
    QueryReturnClass,
    QueryReturnType,
    sql_to_df,
)
from matchbox.common.dtos import (
    DataTypes,
    LocationConfig,
    LocationType,
    Resolution,
    SourceConfig,
    SourceField,
)
from matchbox.common.exceptions import (
    MatchboxResolutionNotFoundError,
    MatchboxSourceClientError,
    MatchboxSourceExtractTransformError,
)
from matchbox.common.graph import ResolutionType
from matchbox.common.hash import HashMethod, hash_rows
from matchbox.common.logging import logger

if TYPE_CHECKING:
    from matchbox.client.dags import DAG
    from matchbox.client.queries import Query

else:
    DAG = Any
    Query = Any


T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")
F = TypeVar("F")


def requires_client(method: Callable[..., T]) -> Callable[..., T]:
    """Decorator that checks if client is set before executing a method.

    A helper method for Location subclasses.

    Raises:
        MatchboxSourceClientError: If the client is not set.
    """

    @wraps(method)
    def wrapper(self: "Location", *args, **kwargs) -> T:
        if self.client is None:
            raise MatchboxSourceClientError
        return method(self, *args, **kwargs)

    return wrapper


class Location(ABC):
    """A location for a data source."""

    def __init__(self, name: str, client: Any):
        """Initialise location."""
        self.config = LocationConfig(type=self.location_type, name=name)
        self.client = client

    def __deepcopy__(self, memo=None):
        """Create a deep copy of the Location object."""
        if memo is None:
            memo = {}

        # Both objects should share the same client
        obj_copy = type(self)(name=deepcopy(self.config.name, memo), client=self.client)

        return obj_copy

    @property
    @abstractmethod
    def location_type(self) -> LocationType:
        """Output location type string."""
        ...

    @abstractmethod
    def connect(self) -> bool:
        """Establish connection to the data location.

        Raises:
            AttributeError: If the client is not set.
        """
        ...

    @abstractmethod
    def validate_extract_transform(self, extract_transform: str) -> bool:
        """Validate ET logic against this location's query language.

        Raises:
            MatchboxSourceExtractTransformError: If the ET logic is invalid.
        """
        ...

    @abstractmethod
    def infer_types(self, extract_transform: str) -> dict[str, DataTypes]:
        """Extract all data types from the ET logic."""
        ...

    @abstractmethod
    def execute(
        self,
        extract_transform: str,
        batch_size: int | None = None,
        rename: dict[str, str] | Callable | None = None,
        return_type: QueryReturnType = QueryReturnType.POLARS,
        keys: tuple[str, list[str]] | None = None,
    ) -> Iterator[QueryReturnClass]:
        """Execute ET logic against location and return batches.

        Args:
            extract_transform: The ET logic to execute.
            batch_size: The size of the batches to return.
            rename: Renaming to apply after the ET logic is executed.

                * If a dictionary is provided, it will be used to rename the columns.
                * If a callable is provided, it will take the old name as input and
                    return the new name.
            return_type: The type of data to return. Defaults to "polars".
            keys: Rule to only retrieve rows by specific keys.
                The key of the dictionary is a field name on which to filter.
                Filters source entries where the key field is in the dict values.

        Raises:
            AttributeError: If the cliet is not set.
        """
        ...

    def from_config(config: LocationConfig, client: Any) -> Self:
        """Initialise location from a location config and an appropriate client."""
        LocClass = location_type_to_class(config.type)
        return LocClass(name=config.name, client=client)


class RelationalDBLocation(Location):
    """A location for a relational database."""

    client: Engine
    location_type: LocationType = LocationType.RDBMS

    @contextmanager
    def _get_connection(self):
        """Context manager for getting database connections with proper cleanup."""
        connection = self.client.connect()
        try:
            yield connection
        finally:
            connection.close()

    @requires_client
    def connect(self) -> bool:  # noqa: D102
        try:
            with self._get_connection() as conn:
                _ = sql_to_df(stmt="select 1", connection=conn)
                return True
        except OperationalError:
            return False

    def validate_extract_transform(self, extract_transform: str) -> bool:  # noqa: D102
        """Check that the SQL statement only contains a single data-extracting command.

        We are NOT attempting a full sanitisation of the SQL statement
        # Validation is done purely to stop accidental mistakes, not malicious actors
        # Users should only run indexing using SourceConfigs they trust and have read,
        # using least privilege credentials

        Args:
            extract_transform: The SQL statement to validate

        Returns:
            bool: True if the SQL statement is valid

        Raises:
            ParseError: If the SQL statement cannot be parsed
            MatchboxSourceExtractTransformError: If validation requirements are not met
        """
        if not extract_transform.strip():
            raise MatchboxSourceExtractTransformError(
                "SQL statement is empty or only contains whitespace."
            )

        expressions = sqlglot.parse(extract_transform)

        if len(expressions) > 1:
            raise MatchboxSourceExtractTransformError(
                "SQL statement contains multiple commands."
            )

        if not expressions:
            raise MatchboxSourceExtractTransformError(
                "SQL statement does not contain any valid expressions."
            )

        expr = expressions[0]

        if not isinstance(expr, sqlglot.expressions.Select):
            raise MatchboxSourceExtractTransformError(
                "SQL statement must start with a SELECT or WITH command."
            )

        forbidden = (
            sqlglot.expressions.DDL,
            sqlglot.expressions.DML,
            sqlglot.expressions.Into,
        )

        if len(list(expr.find_all(forbidden))) > 0:
            raise MatchboxSourceExtractTransformError(
                "SQL statement must not contain DDL or DML commands."
            )

        return True

    @requires_client
    def infer_types(self, extract_transform: str) -> dict[str, DataTypes]:  # noqa: D102
        extract_transform = extract_transform.rstrip(" \t\n;")
        one_row_query = f"select * from ({extract_transform}) as sub limit 1;"
        one_row: pl.DataFrame = list(self.execute(one_row_query))[0]
        column_names = one_row.columns

        inferred_types = {}
        for col in column_names:
            # This expression uses cross-dialect SQL standards;
            # though this is hard to prove as the standard is behind a paywall
            sample_query = (
                f"select {col} from ({extract_transform}) as sub "
                f"where {col} is not null limit 1;"
            )

            sample_row: pl.DataFrame = list(self.execute(sample_query))[0]

            if len(sample_row):
                sample_dtype = sample_row[col].dtype
                inferred_types[col] = DataTypes.from_dtype(sample_dtype)
            else:
                inferred_types[col] = DataTypes.NULL

        return inferred_types

    @requires_client
    def execute(  # noqa: D102
        self,
        extract_transform: str,
        batch_size: int | None = None,
        rename: dict[str, str] | Callable | None = None,
        return_type: QueryReturnType = QueryReturnType.POLARS,
        keys: tuple[str, list[str]] | None = None,
        schema_overrides: dict[str, pl.DataType] | None = None,
    ) -> Generator[QueryReturnClass, None, None]:
        batch_size = batch_size or 10_000
        with self._get_connection() as conn:
            if keys:
                key_field, filter_values = keys
                quoted_key_values = [f"'{key_val}'" for key_val in filter_values]
                # Only filter original SQL if keys are provided
                if quoted_key_values:
                    comma_separated_values = ", ".join(quoted_key_values)
                    # Deal with end-of-statement semi-colons that could be supplied
                    extract_transform = extract_transform.replace(";", "")
                    # This "IN" expression is a SQL standard;
                    # though this is hard to prove as the standard is behind a paywall
                    extract_transform = (
                        f"select * from ({extract_transform}) as sub "
                        f"where {key_field} in ({comma_separated_values})"
                    )
            yield from sql_to_df(
                stmt=extract_transform,
                schema_overrides=schema_overrides,
                connection=conn,
                rename=rename,
                batch_size=batch_size,
                return_batches=True,
                return_type=return_type,
            )


def location_type_to_class(location_type: LocationType) -> type[Location]:
    """Map location type string to the corresponding class."""
    match location_type:
        case LocationType.RDBMS:
            return RelationalDBLocation
        case _:
            raise ValueError("Location type not recognised.")


class Source:
    """Client-side wrapper for source configs."""

    @overload
    def __init__(
        self,
        dag: DAG,
        location: Location,
        name: str,
        extract_transform: str,
        key_field: str,
        index_fields: list[str],
        description: str | None = None,
        infer_types: bool = True,
    ) -> None: ...

    @overload
    def __init__(
        self,
        dag: DAG,
        location: Location,
        name: str,
        extract_transform: str,
        key_field: SourceField,
        index_fields: list[SourceField],
        description: str | None = None,
        infer_types: bool = False,
    ) -> None: ...

    def __init__(
        self,
        dag: DAG,
        location: Location,
        name: str,
        extract_transform: str,
        key_field: str | SourceField,
        index_fields: list[str] | list[SourceField],
        description: str | None = None,
        infer_types: bool = False,
    ):
        """Initialise source.

        Args:
            dag: DAG containing the source.
            location: The location where the source data is stored.
            name: The name of the source.
            description: An optional description of the source.
            extract_transform: The extract/transform logic to apply to the source data.
            key_field: The name of the field to use as the key, or a SourceField
                instance defining the key field. This is the unique identifier we'll
                use to refer to matched data in the source.
            index_fields: The names of the fields to use as index fields, or a list
                of SourceField instances defining the index fields. These are the
                fields you plan to match on.
            infer_types: Whether to infer data types for the fields from the source.
                If False, you must provide SourceField instances for key_field and
                index_fields.
        """
        if not location.validate_extract_transform(extract_transform):
            raise MatchboxSourceExtractTransformError

        self.last_run: datetime | None = None
        self.location = location
        self.dag = dag
        self.name = name
        self.description = description

        if infer_types:
            self._validate_fields(key_field, index_fields, str)

            # Assumes client has been set on location
            inferred_types = location.infer_types(extract_transform)
            remote_fields = {
                field_name: SourceField(name=field_name, type=dtype)
                for field_name, dtype in inferred_types.items()
            }
            typed_key_field = SourceField(name=key_field, type=DataTypes.STRING)
            typed_index_fields = tuple(remote_fields[field] for field in index_fields)
        else:
            typed_key_field, typed_index_fields = self._validate_fields(
                key_field, index_fields, SourceField
            )

        self.config = SourceConfig(
            location_config=location.config,
            extract_transform=extract_transform,
            key_field=typed_key_field,
            index_fields=typed_index_fields,
        )

    def _validate_fields(
        self,
        key_field: Any,
        index_fields: list[Any],
        type_check: type[str] | type[SourceField],
    ) -> tuple[F, tuple[F, ...]]:
        """Validate that fields match the expected type (str or SourceField)."""
        if not isinstance(key_field, type_check):
            raise ValueError(
                f"Expected {type_check.__name__}, got {type(key_field).__name__}"
            )

        if not all(isinstance(f, type_check) for f in index_fields):
            raise ValueError(
                f"All index_fields must be {type_check.__name__} instances"
            )

        return key_field, tuple(index_fields)

    def to_resolution(self) -> Resolution:
        """Convert to Resolution for API calls."""
        return Resolution(
            name=self.name,
            description=self.description,
            truth=None,
            resolution_type=ResolutionType.SOURCE,
            config=self.config,
        )

    @classmethod
    def from_resolution(
        cls, resolution: Resolution, dag: DAG, location: Location
    ) -> "Source":
        """Reconstruct from Resolution."""
        if resolution.resolution_type != ResolutionType.SOURCE:
            raise ValueError("Resolution must be of type 'source'")

        return cls(
            dag=dag,
            location=location,
            name=resolution.name,
            extract_transform=resolution.config.extract_transform,
            key_field=resolution.config.key_field,
            index_fields=resolution.config.index_fields,
            description=resolution.description,
            infer_types=False,
        )

    def __hash__(self) -> int:
        """Return a hash of the Source based on its config."""
        return hash(self.config)

    def __eq__(self, other: Any) -> bool:
        """Check equality of two Source objects based on their config."""
        if not isinstance(other, Source):
            return False
        return self.config == other.config

    def fetch(
        self,
        qualify_names: bool = False,
        batch_size: int | None = None,
        return_type: QueryReturnType = QueryReturnType.POLARS,
        keys: list[str] | None = None,
    ) -> Generator[QueryReturnClass, None, None]:
        """Applies the extract/transform logic to the source and returns the results.

        Args:
            qualify_names: If True, qualify the names of the columns with the
                source name.
            batch_size: Indicate the size of each batch when fetching data in batches.
            return_type: The type of data to return. Defaults to "polars".
            keys: List of keys to select a subset of all source entries.

        Returns:
            The requested data in the specified format, as an iterator of tables.
        """
        _rename: Callable | None = None
        if qualify_names:

            def _rename(c: str) -> str:
                return self.name + "_" + c

        all_fields = self.config.index_fields + tuple([self.config.key_field])
        schema_overrides = {field.name: field.type.to_dtype() for field in all_fields}

        if keys:
            yield from self.location.execute(
                extract_transform=self.config.extract_transform,
                schema_overrides=schema_overrides,
                rename=_rename,
                batch_size=batch_size,
                return_type=return_type,
                keys=(self.config.key_field.name, keys),
            )
        else:
            yield from self.location.execute(
                extract_transform=self.config.extract_transform,
                schema_overrides=schema_overrides,
                rename=_rename,
                batch_size=batch_size,
                return_type=return_type,
            )

    def run(
        self, batch_size: int | None = None, full_rerun: bool = False
    ) -> ArrowTable:
        """Hash a dataset from its warehouse, ready to be inserted, and cache hashes.

        Hashes the index fields defined in the source based on the
        extract/transform logic.

        Does not hash the key field.

        Args:
            batch_size: If set, process data in batches internally. Indicates the
                size of each batch.
            full_rerun: Whether to force a re-run even if the hashes are cached


        Returns:
            A PyArrow Table containing source keys and their hashes.
        """
        if self.last_run and not full_rerun:
            warnings.warn("Source already run, skipping.", UserWarning, stacklevel=2)
            return self.hashes

        log_prefix = f"Hash {self.name}"
        batch_info = (
            f"with batch size {batch_size:,}" if batch_size else "without batching"
        )
        logger.debug(f"Retrieving and hashing {batch_info}", prefix=log_prefix)

        key_field: str = self.config.key_field.name
        index_fields: list[str] = [field.name for field in self.config.index_fields]

        all_results: list[pl.DataFrame] = []
        for batch in self.fetch(
            batch_size=batch_size,
            return_type="polars",
        ):
            batch: pl.DataFrame

            if batch[key_field].is_null().any():
                raise ValueError("keys column contains null values")

            row_hashes: pl.Series = hash_rows(
                df=batch,
                columns=list(sorted(index_fields)),
                method=HashMethod.SHA256,
            )

            result = (
                batch.rename({key_field: "keys"})
                .with_columns(row_hashes.alias("hash"))
                .select(["hash", "keys"])
            )
            all_results.append(result)

        self.hashes = (
            pl.concat(all_results).group_by("hash").agg(pl.col("keys")).to_arrow()
        )

        self.last_run = datetime.now()

        return self.hashes

    # Note: name, description, truth are now instance variables, not properties

    @property
    def prefix(self) -> str:
        """Get the prefix for the source."""
        return self.config.prefix(self.name)

    @property
    def qualified_key(self) -> str:
        """Get the qualified key for the source."""
        return self.config.qualified_key(self.name)

    @property
    def qualified_index_fields(self) -> list[str]:
        """Get the qualified index fields for the source."""
        return self.config.qualified_index_fields(self.name)

    def qualify_field(self, field: str) -> str:
        """Qualify field names with the source name.

        Args:
            field: The field name to qualify.

        Returns:
            A single qualified field.

        """
        return self.config.qualify_field(self.name, field)

    def f(self, fields: str | Iterable[str]) -> str | list[str]:
        """Qualify one or more field names with the source name.

        Args:
            fields: The field name to qualify, or a list of field names.

        Returns:
            A single qualified field, or a list of qualified field names.

        """
        return self.config.f(self.name, fields)

    def sync(self) -> None:
        """Send the source config and hashes to the server."""
        resolution = self.to_resolution()
        try:
            existing_resolution = _handler.get_resolution(name=self.name)
        except MatchboxResolutionNotFoundError:
            existing_resolution = None
        # Check if config matches
        if existing_resolution:
            if existing_resolution.config != self.config:
                raise ValueError(
                    f"Resolution {self.name} already exists with different "
                    "configuration. Please delete the existing resolution "
                    "or use a different name. "
                )
            else:
                log_prefix = f"Resolution {self.name}"
                logger.warning("Already exists. Passing.", prefix=log_prefix)
        else:
            _handler.create_resolution(resolution=resolution)

        if self.hashes:
            _handler.set_data(
                name=self.name, data=self.hashes, validate_type=ResolutionType.SOURCE
            )

    def query(self, **kwargs) -> Query:
        """Generate a query for this source."""
        return self.dag.query(self, **kwargs)
